"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import shutil
import os

from . import transphire_utils as tu


def create_substack_command(class_average_name, input_stack, isac_dir, output_dir, settings):
    command = []
    block_gpu = False
    gpu_list = []
    shell = True
    check_files = []
    gpu_list = []

    try:
        shutil.rmtree(output_dir)
    except IOError:
        pass

    command.append("PATH=$(dirname $(which {0})):${{PATH}}".format(settings['Path']['sp_pipe.py']))
    command.append(settings['Path']['sp_pipe.py'])
    command.append('isac_substack')
    command.append(input_stack)
    command.append(os.path.join(isac_dir, 'ISAC2'))
    command.append(os.path.join(output_dir, 'STACK'))
    command.append('--isac_class_avgs_path={0}'.format(class_average_name))

    #picking_name = settings['Copy']['Picking']
    #if settings[picking_name]['--filament'] == 'True':
    #    command.append('--min_nr_segments={0}'.format(settings[picking_name]['--minimum_number_boxes']))

    return ' '.join(command), check_files, block_gpu, gpu_list, shell, 'bdb:{0}/{1}'.format(os.path.join(output_dir, 'STACK'), 'isac_substack')


def create_restack_command(stack_name, output_dir, settings):
    prog_name = settings['Copy']['Train2d']
    command = []
    block_gpu = False
    gpu_list = []
    shell = True
    check_files = []
    gpu_list = []

    command.append("PATH=$(dirname $(which {0})):${{PATH}}".format(settings['Path']['sp_pipe.py']))
    command.append(settings['Path']['sp_pipe.py'])
    command.append('restacking')
    command.append(stack_name)
    command.append(os.path.join(output_dir, 'BOX'))
    command.append('--rb_box_size={0}'.format(settings[prog_name]['Box size']))
    command.append('--reboxing')

    return ' '.join(command), check_files, block_gpu, gpu_list, shell, os.path.join(output_dir, 'BOX')


def create_eval_command(config_file, weight_file, log_file, settings, name):
    prog_name = settings['Path']['cryolo_evaluation.py']
    command = []
    block_gpu = False
    gpu_list = []
    shell = False
    check_files = []
    gpu_list = []

    with open(log_file, 'r') as read:
        match = re.search('^.*Wrote runfile to: (.*)$', read.read(), re.M) # https://regex101.com/r/BY0m3B/1

    if match is None:
        return None, check_files, block_gpu, gpu_list, shell

    command.append(prog_name)
    command.append('--config={0}'.format(config_file))
    command.append('--weights={0}'.format(weight_file))
    command.append('--runfile={0}'.format(match.group(1).strip()))

    prog_name_train = settings['Copy']['Train2d']
    if settings[prog_name_train]['Split Gpu?'] == 'True':
        try:
            gpu_id = int(name.split('_')[-1])-1
        except ValueError:
            gpu_id = 0
        try:
            gpu = settings[prog_name_train]['--gpu'].split()[gpu_id]
        except IndexError:
            raise UserWarning('There are less gpus provided than threads available! Please restart with the same number of pipeline processors as GPUs provided and restart! Stopping this thread!')
    else:
        gpu = settings[prog_name_train]['--gpu']
    command.append('--gpu={0}'.format(gpu))

    return ' '.join(command), check_files, block_gpu, gpu.split(), shell


def create_train_command(sum_folder, box_folder, output_dir, name, settings):
    prog_name = settings['Copy']['Train2d']
    picking_name = settings['Copy']['Picking']
    command = []
    block_gpu = False
    shell = True
    check_files = []

    config_file = os.path.join(output_dir, 'config.json')
    weight_file = os.path.join(output_dir, 'weight.h5')

    command.append("PATH=$(dirname $(which {0})):${{PATH}}".format(settings['Path']['cryolo_gui.py']))
    command.append(settings['Path']['cryolo_gui.py'])
    command.append('config')
    command.append(config_file)
    command.append(settings[prog_name]['Box size'])
    command.append('--train_image_folder={0}'.format(sum_folder))
    command.append('--train_annot_folder={0}'.format(box_folder))
    command.append('--saved_weights_name={0}'.format(weight_file))
    command.append('--pretrained_weights={0}'.format(settings[picking_name]['--weights_old']))
    command.append('--filtered_output={0}'.format(os.path.join(
        settings['project_folder'],
        'XXX_Filtered_Images'
        )))
    command.append('--train_times={0}'.format(settings[prog_name]['--train_times']))

    command.append(';')
    command.append("PATH=$(dirname $(which {0})):${{PATH}}".format(settings['Path'][prog_name]))
    command.append(settings['Path'][prog_name])
    command.append('-c {0}'.format(config_file))

    ignore_list = []
    ignore_list.append('--fine_tune')
    ignore_list.append('--use_multithreading')
    for entry in ignore_list:
        try:
            if settings[prog_name][entry] == 'True':
                command.append(entry)
        except KeyError:
            pass

    if tu.is_higher_version(prog_name, '1.7.4'):
        try:
            command.remove('--use_multithreading')
        except ValueError:
            pass

    ignore_list.append('Split Gpu?')
    ignore_list.append('--gpu')
    ignore_list.append('Box size')
    ignore_list.append('--train_times')
    ignore_list.append('Maximum micrographs')
    ignore_list.append('Use centered')

    if settings[prog_name]['Split Gpu?'] == 'True':
        try:
            gpu_id = int(name.split('_')[-1])-1
        except ValueError:
            gpu_id = 0
        try:
            gpu = settings[prog_name]['--gpu'].split()[gpu_id]
        except IndexError:
            raise UserWarning('There are less gpus provided than threads available! Please restart with the same number of pipeline processors as GPUs provided and restart! Stopping this thread!')
    else:
        gpu = settings[prog_name]['--gpu']

    command.append('--gpu')
    command.append('{0}'.format(gpu))

    for key in settings[prog_name]:
        if key in ignore_list:
            continue
        else:
            command.append(key)
            command.append(
                '{0}'.format(settings[prog_name][key])
                )

    return ' '.join(command), check_files, block_gpu, gpu.split(), shell, weight_file, config_file



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QLabel, QPushButton, QDialogButtonBox, QWidget


class MessageBox(QDialog):
    """
    Show a message box

    Inherits:
    QDialog

    Signals:
    None
    """

    def __init__(self, is_question, parent=None):
        """
        Initialise layout of the widget.

        Arguments:
        folder - Folder name to mount.
        default - Default user name
        login - True, if it is possible to test the users credentials on this machine
        extension - True, if a mount extension is required
        parent - Parent widget

        Returns:
        None
        """
        super(MessageBox, self).__init__(parent)
        central_raw_layout = QVBoxLayout(self)
        central_raw_layout.setContentsMargins(0, 0, 0, 0)
        central_widget_raw = QWidget(self)
        central_widget_raw.setObjectName('central_raw')
        central_raw_layout.addWidget(central_widget_raw)

        central_layout = QVBoxLayout(central_widget_raw)
        central_widget = QWidget(self)
        central_widget.setObjectName('central')
        central_layout.addWidget(central_widget)

        layout = QVBoxLayout(central_widget)
        self.label = QLabel(self)

        button_box = QDialogButtonBox(self)
        if is_question:
            accept_button = QPushButton('Yes')
            accept_button.setObjectName('start')
            reject_button = QPushButton('No')
            reject_button.setObjectName('stop')

            button_box.addButton(accept_button, QDialogButtonBox.AcceptRole)
            button_box.addButton(reject_button, QDialogButtonBox.RejectRole)

            button_box.accepted.connect(self.accept)
            button_box.rejected.connect(self.reject)

        else:
            accept_button = QPushButton('Ok')
            accept_button.setObjectName('start')

            button_box.addButton(accept_button, QDialogButtonBox.AcceptRole)

            button_box.accepted.connect(self.accept)

        layout.addWidget(self.label)
        layout.addWidget(button_box)

    def setText(self, heading, text):
        """
        Set the text to the label.

        Arguments:
        heading - Heading of the window
        text - Text of the label

        Returns:
        None
        """
        if heading is not None:
            self.setWindowTitle(heading)
        self.label.setText(text)

    def setDefault(self, text):
        pass



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
import os
#import re
import collections as co
import numpy as np
import mrcfile as mrc
import matplotlib
matplotlib.use('QT5Agg')
import matplotlib.image as mi

from . import transphire_utils as tu
from . import transphire_software as ts


def get_motion_default(settings, motion_frames, queue_com, name):
    """
    Set the default values for the motion correction software.

    settings - TranSPHIRE settings.
    motion_frames - Sub frame settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    True, if dose weighting will be applied.
    """

    motion_name = settings['Copy']['Motion']
    if motion_name.startswith('MotionCor2'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            if motion_frames is not None:
                motion_frames['last'] = \
                    int(settings['Input']['Number of frames']) - \
                    int(settings[motion_name]['-Trunc'])
                motion_frames['first'] = \
                    int(settings[motion_name]['-Throw']) + 1

            return bool(
                    settings[motion_name]['-FmDose'] != '0' and
                    settings[motion_name]['-PixSize'] != '0' and
                    settings[motion_name]['-kV'] != '0'
                    )

        else:
            message = '\n'.join([
                '{0}: Motion version not known.'.format(motion_name),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
    elif motion_name.startswith('Unblur'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            return bool(
                    settings[motion_name]['Pixel size of image (A)'] != '0' and
                    settings[motion_name]['Exposure per frame (e/A^2)'] != '0' and
                    settings[motion_name]['Acceleration voltage'] != '0'
                    )

        else:
            message = '\n'.join([
                '{0}: Motion version not known.'.format(motion_name),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
    else:
        message = '\n'.join([
            '{0}: Motion name not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
    raise IOError(message)


def get_dws_file_name(output_transfer_scratch, file_name, settings, queue_com, name):
    """
    Get the name of the dose weighted file directly after the program finished.

    output_transfer - Name of the folder in the scratch directory.
    file_name - File name of the root_name path.
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    File path of the DW file.
    """
    motion_name = settings['Copy']['Motion']
    if motion_name.startswith('MotionCor2'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            return os.path.join(
                output_transfer_scratch,
                '{0}_DWS.mrc'.format(file_name)
                )

        else:
            message = '\n'.join([
                '{0}: Motion version not known.'.format(motion_name),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
            raise IOError(message)
    elif motion_name.startswith('Unblur'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            return 'None_TRANSPHIRE_DUMMY'

        else:
            message = '\n'.join([
                '{0}: Motion version not known.'.format(motion_name),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
            raise IOError(message)


def get_dw_file_name(output_transfer_scratch, file_name, settings, queue_com, name):
    """
    Get the name of the dose weighted file directly after the program finished.

    output_transfer - Name of the folder in the scratch directory.
    file_name - File name of the root_name path.
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    File path of the DW file.
    """
    motion_name = settings['Copy']['Motion']
    if motion_name.startswith('MotionCor2'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            return os.path.join(
                output_transfer_scratch,
                '{0}_DW.mrc'.format(file_name)
                )

        else:
            message = '\n'.join([
                '{0}: Motion version not known.'.format(motion_name),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
            raise IOError(message)
    elif motion_name.startswith('Unblur'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            return os.path.join(
                output_transfer_scratch,
                '{0}_DW.mrc'.format(file_name)
                )

        else:
            message = '\n'.join([
                '{0}: Motion version not known.'.format(motion_name),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
            raise IOError(message)


def get_motion_command(file_input, file_output_scratch, file_log_scratch, settings, queue_com, name, set_name, do_subsum):
    """
    Get the command for the selected motion software.

    file_input - Input file for motion correction.
    file_output_scratch - Output file name
    file_log_scratch - Logfile path on the scratch disc
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    Motion command
    """
    motion_name = settings['Copy']['Motion']
    shell = False
    command = None
    block_gpu = None
    gpu_list = None
    file_to_delete = None
    if motion_name.startswith('MotionCor2'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            mem_is_one = bool(float(settings[motion_name]['-GpuMemUsage']) == 1)
            if mem_is_one:
                settings[motion_name]['-GpuMemUsage'] = '0.5'
            command, gpu = create_motion_cor_2_v1_0_0_command(
                motion_name=motion_name,
                file_input=file_input,
                file_output=file_output_scratch,
                file_log=file_log_scratch,
                settings=settings,
                queue_com=queue_com,
                name=name,
                set_name=set_name,
                do_subsum=do_subsum,
                )
            gpu_list = gpu.split()

            if tu.is_higher_version(motion_name, '1.2.0'):
                if '_' in gpu:
                    if mem_is_one:
                        raise UserWarning('Sub GPUs are only supported if the GpuMemUsage option is not equal 1')
                    block_gpu = True
                    file_to_delete = '/tmp/MotionCor2_FreeGpus.txt'
                elif float(settings[motion_name]['-GpuMemUsage']) == 0:
                    block_gpu = False
                    file_to_delete = '/tmp/MotionCor2_FreeGpus.txt'
                else:
                    block_gpu = True
            elif tu.is_higher_version(motion_name, '1.1.0'):
                if '_' in gpu:
                    if mem_is_one:
                        raise UserWarning('Sub GPUs are only supported if the GpuMemUsage option is not equal 1')
                    block_gpu = True
                elif float(settings[motion_name]['-GpuMemUsage']) == 0:
                    block_gpu = False
                else:
                    block_gpu = True
            elif tu.is_higher_version(motion_name, '1.0.5'):
                block_gpu = True
                if '_' in gpu:
                    raise UserWarning('Sub GPUs are only supported in MotionCor version >=1.1.0, <1.2.0')
            elif tu.is_higher_version(motion_name, '1.0.0'):
                block_gpu = False
                if '_' in gpu:
                    raise UserWarning('Sub GPUs are only supported in MotionCor version >=1.1.0, <1.2.0')

        else:
            message = '\n'.join([
                '{0}: Motion version not known.'.format(settings['Copy']['Motion']),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
            raise IOError(message)
    elif motion_name.startswith('Unblur'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            gpu_list = []
            block_gpu = False
            shell = True
            command = create_unblur_v1_0_0_command(
                motion_name=motion_name,
                file_input=file_input,
                file_output=file_output_scratch,
                file_log=file_log_scratch,
                settings=settings,
                queue_com=queue_com,
                name=name,
                set_name=set_name,
                do_subsum=do_subsum,
                )

    assert command is not None, 'command not specified: {0}'.format(motion_name)
    assert block_gpu is not None, 'block_gpu not specified: {0}'.format(motion_name)
    assert gpu_list is not None, 'gpu_list not specified: {0}'.format(motion_name)

    return command, block_gpu, gpu_list, file_to_delete, shell


def create_unblur_v1_0_0_command(
        motion_name,
        file_input,
        file_output,
        file_log,
        settings,
        queue_com,
        name,
        set_name,
        do_subsum,
        ):
    """
    Create the SumMovie v1.0.2 command.

    motion_frames - Sub frames settings dictionary
    file_input - File to sum.
    file_output - Output file name
    file_shift - Output shift file name
    file_frc - Output frc file name
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of the process

    Returns:
    Command for Summovie v1.0.2
    """
    do_dw = get_motion_default(
        settings,
        None,
        queue_com,
        name)


    commands = []

    for weighting_entry in [0] + [1] * do_dw:
        if weighting_entry:
            file_output_use = '{}_DW{}'.format(*os.path.splitext(file_output))
        else:
            file_output_use = file_output
        cmd = []
        # Input stack filename
        cmd.append(file_input)
        # Output aligned sum
        cmd.append(file_output_use)
        cmd.append(settings[motion_name]['Pixel size of image (A)'])
        cmd.append(settings[motion_name]['Output binning factor'])
        if weighting_entry:
            cmd.append('yes')
            cmd.append(settings[motion_name]['Acceleration voltage'])
            cmd.append(settings[motion_name]['Exposure per frame (e/A^2)'])
            cmd.append(settings[motion_name]['Pre-exposure amount (e/A^2)'])
        else:
            cmd.append('no')
        cmd.append('yes')
        cmd.append(settings[motion_name]['Minimum shift for initial search (A)'])
        cmd.append(settings[motion_name]['Outer radius shift limit (A)'])
        cmd.append(settings[motion_name]['B-factor to apply to images (A^2)'])
        cmd.append(settings[motion_name]['Half-width of vertical Fourier mask'])
        cmd.append(settings[motion_name]['Half-width of horizontal Fourier mask'])
        cmd.append(settings[motion_name]['Termination shift threshold (A)'])
        cmd.append(settings[motion_name]['Maximum number of iterations'])
        if weighting_entry:
            if settings[motion_name]['Restore Noise Power?'] == 'True':
                cmd.append('yes')
            else:
                cmd.append('no')
        if settings[motion_name]['Gain image filename'] != '':
            cmd.append('no')
            external_log, key = settings[motion_name]['Gain image filename'].split('|||')
            with open(settings[external_log], 'r') as read:
                log_data = json.load(read)
            gain_file = log_data[set_name][motion_name][key]['new_file']
            cmd.append(gain_file)
        else:
            cmd.append('yes')
        cmd.append(settings[motion_name]['First frame to use for sum'])
        cmd.append(settings[motion_name]['Last frame to use for sum'])
        if settings[motion_name]['Correct Magnification Distortion?'] == 'True':
            cmd.append('yes')
            cmd.append(settings[motion_name]['Distortion Angle (Degrees)'])
            cmd.append(settings[motion_name]['Major Scale'])
            cmd.append(settings[motion_name]['Minor Scale'])
        else:
            cmd.append('no')


        commands.append('echo "{0}" | {1}'.format(
            '\n'.join(cmd),
            '{0}'.format(settings['Path'][motion_name])
            ))

    return ';'.join(commands)


def create_motion_cor_2_v1_0_0_command(motion_name, file_input, file_output, file_log, settings, queue_com, name, set_name, do_subsum):
    """
    Create the MotionCor2 v1.0.0 command

    file_input - Input file for motion correction.
    file_output - Output filename
    file_log - Logfile name
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    Command for MotionCor2 v1.0.0
    """

    ignore_list = []
    ignore_list.append('Split Gpu?')
    ignore_list.append('-Gpu')
    ignore_list.append('dose cutoff')
    command = []
    # Start the program
    command.append('{0}'.format(settings['Path'][motion_name]))
    # Input Micrograph
    _, extension = os.path.splitext(file_input)
    if extension == '.tiff' or \
            extension == '.tif':
        command.append('-InTiff')
        command.append('{0}'.format(file_input))

    elif extension == '.mrc':
        command.append('-InMrc')
        command.append('{0}'.format(file_input))

    else:
        message = '{0}: Not known!'.format(extension)
        queue_com['error'].put(message, name)
        raise IOError(message)

    # Output micrograph
    command.append('-OutMrc')
    command.append('{0}'.format(file_output))
    if do_subsum:
        # Write the output stack
        command.append('-OutStack')
        command.append('1')
    else:
        pass
    # Log file
    command.append('-LogFile')
    command.append('{0}'.format(file_log))

    if settings[motion_name]['Split Gpu?'] == 'True':
        try:
            gpu_id = int(name.split('_')[-1])
        except ValueError:
            gpu_id = 0
        try:
            gpu_raw = settings[motion_name]['-Gpu'].split()[gpu_id]
        except IndexError:
            raise UserWarning('There are less gpus provided than threads available! Please restart with the same number of pipeline processors as GPUs provided and restart! Stopping this thread!')
    else:
        gpu_raw = settings[motion_name]['-Gpu']

    gpu = ' '.join(list(set([entry.split('_')[0] for entry in gpu_raw.split()])))
    if len(gpu.split()) != len(gpu_raw.split()) and settings[motion_name]['Split Gpu?'] == 'False':
        raise UserWarning('One cannot use multi GPU in combination with the disabled Split GPU option!')

    command.append('-Gpu')
    command.append('{0}'.format(gpu))

    for key, value in settings[motion_name].items():
        if key in ignore_list:
            continue
        elif value:
            try:
                if '|||' in value:
                    external_log, local_key = value.split('|||')
                    with open(settings[external_log], 'r') as read:
                        log_data = json.load(read)
                    try:
                        set_value = log_data[set_name][motion_name][local_key]['new_file']
                    except KeyError:
                        continue
                else:
                    set_value = value
            except TypeError:
                set_value = value

            command.append(key)
            command.append('{0}'.format(set_value))
        else:
            continue

    return ' '.join(command), gpu_raw


def create_sum_movie_command(
        motion_frames, file_input, file_output, file_shift, file_frc,
        settings, queue_com, name
        ):
    """
    Create the SumMovie command.

    motion_frames - Sub frames settings dictionary
    file_input - File to sum.
    file_output - Output file name
    file_shift - Output shift file name
    file_frc - Output frc file name
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of the process

    Returns:
    Command for SumMovie
    """
    command = create_sum_movie_v1_0_2_command(
        motion_frames=motion_frames,
        file_input=file_input,
        file_output=file_output,
        file_shift=file_shift,
        file_frc=file_frc,
        settings=settings,
        queue_com=queue_com,
        name=name
        )
    block_gpu = False
    gpu_list = []
    return command, block_gpu, gpu_list


def create_sum_movie_v1_0_2_command(
        motion_frames, file_input, file_output, file_shift, file_frc,
        settings, queue_com, name
        ):
    """
    Create the SumMovie v1.0.2 command.

    motion_frames - Sub frames settings dictionary
    file_input - File to sum.
    file_output - Output file name
    file_shift - Output shift file name
    file_frc - Output frc file name
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of the process

    Returns:
    Command for Summovie v1.0.2
    """
    sum_movie_command = []
    # Input file
    sum_movie_command.append('{0}'.format(file_input))
    # Number of frames

    motion_name = settings['Copy']['Motion']
    if tu.is_higher_version(motion_name, '1.0.0'):
        sum_movie_command.append('{0}'.format(
            int(settings['Input']['Number of frames']) -
            int(settings[motion_name]['-Trunc']) -
            int(settings[motion_name]['-Throw'])
            ))

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    # Output sum file
    sum_movie_command.append('{0}'.format(file_output))
    # Shift file
    sum_movie_command.append('{0}'.format(file_shift))
    # FRC file
    sum_movie_command.append('{0}'.format(file_frc))
    # First frame
    sum_movie_command.append('{0}'.format(motion_frames['first']))
    # Last frame
    sum_movie_command.append('{0}'.format(motion_frames['last']))
    # Pixel size

    if tu.is_higher_version(motion_name, '1.0.0'):
        sum_movie_command.append(
            '{0}'.format(settings[motion_name]['-PixSize'])
            )

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    # Dose correction
    sum_movie_command.append('No')

    command = 'echo "{0}" | {1}'.format(
        '\n'.join(sum_movie_command),
        '{0}'.format(settings['Path']['SumMovie v1.0.2'])
        )

    return command


def combine_motion_outputs(
        data,
        data_original,
        settings,
        queue_com,
        shared_dict,
        name,
        log_file,
        sum_file,
        dw_file,
        stack_file,
        set_name,
        ):
    """
    Combine the motion outputs to one micrograph and one relion star file.

    root_path - Root path of the file
    file_name - File name of the ctf file.
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of process
    sum_file - Name of the dose uncorrected sum file

    Returns:
    None
    """
    motion_name = settings['Copy']['Motion']
    motion_settings = settings[motion_name]
    log_folder = os.path.dirname(log_file)
    stack_folder = settings['stack_folder']
    compress_folder = settings['compress_folder_feedback_0']
    project_folder = '{0}/'.format(settings['project_folder'].rstrip('/'))
    project_base = '{0}/'.format(settings['project_base'].rstrip('/'))
    sum_file = sum_file.replace(project_base, '')
    if dw_file:
        dw_file = dw_file.replace(project_base, '')
    else:
        pass

    output_name_mic = os.path.join(
        log_folder,
        '{0}_transphire_motion.txt'.format(os.path.basename(sum_file))
        )
    output_name_star = os.path.join(
        log_folder,
        '{0}_transphire_motion.star'.format(os.path.basename(sum_file))
        )
    output_name_star_relion3 = os.path.join(
        log_folder,
        '{0}_transphire_motion_relion3.star'.format(os.path.basename(sum_file))
        )

    output_name_mic_combined = os.path.join(
        project_folder,
        '{0}_transphire_motion.txt'.format(motion_name.replace(' ', '_').replace('>=', ''))
        )
    output_name_star_combined = os.path.join(
        project_folder,
        '{0}_transphire_motion.star'.format(motion_name.replace(' ', '_').replace('>=', ''))
        )
    output_name_star_relion3_combined = os.path.join(
        project_folder,
        '{0}_transphire_motion_relion3.star'.format(motion_name.replace(' ', '_').replace('>=', ''))
        )

    # SPHIRE
    with open(output_name_mic, 'w') as write:
        write.write('{0}\n'.format(os.path.basename(sum_file)))

    # RELION 2
    header_star = co.OrderedDict()
    if dw_file:
        header_star['_rlnMicrographNameNoDW'] = sum_file
        header_star['_rlnMicrographName'] = dw_file
    else:
        header_star['_rlnMicrographName'] = sum_file

    export_lines_star = [get_relion_header(header_star.keys())]
    create_export_data(header_star.values(), export_lines_star)
    with open(output_name_star, 'w') as write:
        write.write('\n'.join(export_lines_star))

    # RELION 3
    #with open(log_file, 'r') as read:
    #    data_read = read.read()

    var_dict = {
        'size_x': None,
        'size_y': None,
        'size_z': None,
        'gain_var': None,
        'defect_var': None,
        'bin_var': None,
        'pixel_size_var': None,
        'dose_var': None,
        'init_dose_var': None,
        'voltage_var': None,
        'throw_var': None,
        }

    _, var_dict['size_x'], var_dict['size_y'], var_dict['size_z'] = ts.check_nr_frames([stack_file], settings, force=True)
    movie_name = stack_file
    if motion_name.startswith('MotionCor2'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            var_dict['gain_var'] = motion_settings['-Gain']
            var_dict['defect_var'] = motion_settings['-DefectFile']
            var_dict['bin_var'] = motion_settings['-FtBin']
            var_dict['pixel_size_var'] = motion_settings['-PixSize']
            var_dict['dose_var'] = motion_settings['-FmDose']
            var_dict['init_dose_var'] = motion_settings['-InitDose']
            var_dict['voltage_var'] = motion_settings['-kV']
            var_dict['throw_var'] = int(motion_settings['-Throw']) + 1

    elif motion_name.startswith('Unblur'):
        if tu.is_higher_version(motion_name, '1.0.0'):
            var_dict['gain_var'] = motion_settings['Gain image filename']
            var_dict['defect_var'] = False
            var_dict['bin_var'] = motion_settings['Output binning factor']
            var_dict['pixel_size_var'] = motion_settings['Pixel size of image (A)']
            var_dict['dose_var'] = motion_settings['Exposure per frame (e/A^2)']
            var_dict['init_dose_var'] = motion_settings['Pre-exposure amount (e/A^2)']
            var_dict['voltage_var'] = motion_settings['Acceleration voltage']
            var_dict['throw_var'] = int(motion_settings['First frame to use for sum'])

    for key, value in var_dict.items():
        if value is None:
            assert False, (var_dict, key, value)

        try:
            external_log, json_key = value.split('|||')
        except ValueError:
            continue
        except AttributeError:
            continue
        with open(settings[external_log], 'r') as read:
            log_data = json.load(read)
        var_dict[key] = log_data[set_name][motion_name][json_key]['new_file']

    compress_name = settings['Copy']['Compress']
    if compress_name == 'False' or settings['Input']['Input frames extension'] in ('tiff', 'tif'):
        pass
    elif compress_name == 'Compress cmd':
        movie_name = movie_name.replace(stack_folder, compress_folder)
        movie_name = movie_name.replace(
            '.' + settings['Input']['Input frames extension'],
            '.' + settings[compress_name]['--command_compress_extension']
            )
    else:
        assert False, compress_name

    data_meta = [
        '',
        'data_general',
        '',
        '_rlnImageSizeX {0}'.format(var_dict['size_x']),
        '_rlnImageSizeY {0}'.format(var_dict['size_y']),
        '_rlnImageSizeZ {0}'.format(var_dict['size_z']),
        '_rlnMicrographMovieName {0}'.format(movie_name.replace(project_base, '')),
        ]
    if var_dict['gain_var']:
        new_gain = var_dict['gain_var']
        data_meta.extend([
            '_rlnMicrographGainName {0}'.format(new_gain.replace(project_base, '')),
            ])
    else:
        new_gain = None

    if var_dict['defect_var']:
        new_defect = var_dict['defect_var']
        data_meta.extend([
            '_rlnMicrographDefectFile {0}'.format(new_defect.replace(project_base, '')),
            ])
    else:
        new_defect = None

    data_meta.extend([
        '_rlnMicrographBinning {0}'.format(var_dict['bin_var']),
        ])
    data_meta.extend([
        '_rlnMicrographOriginalPixelSize {0}'.format(var_dict['pixel_size_var']),
        ])
    data_meta.extend([
        '_rlnMicrographDoseRate {0}'.format(var_dict['dose_var']),
        ])
    data_meta.extend([
        '_rlnMicrographPreExposure {0}'.format(var_dict['init_dose_var']),
        ])
    data_meta.extend([
        '_rlnVoltage {0}'.format(var_dict['voltage_var']),
        ])

    data_meta.extend([
        '_rlnMicrographStartFrame {0}'.format(var_dict['throw_var']),
        '_rlnMotionModelVersion 0',
        '',
        'data_global_shift',
        '',
        'loop_',
        '_rlnMicrographFrameNumber #1',
        '_rlnMicrographShiftX #2',
        '_rlnMicrographShiftY #3',
        ])

    assert data_original.shape[0] == 1, data_original
    data_original = data_original[0]
    idx_x = 0
    idx_y = 1

    sum_total = 0
    sum_early = 0
    sum_late = 0
    frame_cutoff = float(motion_settings['dose cutoff']) - float(var_dict['init_dose_var'])
    try:
        frame_cutoff /= float(var_dict['dose_var'])
    except ZeroDivisionError:
        data_meta = ['-FmDose needs to be specified in order to use the meta files']
    else:
        for idx in range(1, data_original.shape[1]):
            drift = np.sqrt(
                (data_original[idx_x][idx-1] - data_original[idx_x][idx])**2 +
                (data_original[idx_y][idx-1] - data_original[idx_y][idx])**2
                )
            sum_total += drift
            if idx <= frame_cutoff:
                sum_early += drift
            else:
                sum_late += drift

        for idx in range(data_original.shape[1]):
            data_meta.extend(['{0}\t{1}\t{2}'.format(
                idx+1,
                data_original[idx_x][idx],
                data_original[idx_y][idx]
                )])

    relion3_meta = os.path.join(
        log_folder,
        '{0}_transphire_motion_relion3_meta.star'.format(os.path.basename(sum_file))
        )
    with open(relion3_meta, 'w') as write:
        write.write('\n'.join(data_meta))

    header_star_relion3 = co.OrderedDict()
    if dw_file:
        header_star_relion3['_rlnMicrographNameNoDW'] = sum_file.replace(project_base, '')
        header_star_relion3['_rlnMicrographName'] = dw_file.replace(project_base, '')
    else:
        header_star_relion3['_rlnMicrographName'] = sum_file.replace(project_base, '')
    header_star_relion3['_rlnMicrographMetadata'] = relion3_meta.replace(project_base, '')
    if sum_total:
        header_star_relion3['_rlnAccumMotionTotal'] = sum_total
        header_star_relion3['_rlnAccumMotionEarly'] = sum_early
        header_star_relion3['_rlnAccumMotionLate'] = sum_late
    else:
        pass

    export_lines_star_relion3 = [get_relion_header(header_star_relion3.keys())]
    create_export_data(header_star_relion3.values(), export_lines_star_relion3)

    with open(output_name_star_relion3, 'w') as write:
        write.write('\n'.join(export_lines_star_relion3))

    return (
        output_name_mic_combined,
        output_name_star_combined,
        output_name_star_relion3_combined,
        output_name_mic,
        output_name_star,
        output_name_star_relion3,
        relion3_meta
        )


def get_relion_header(names):
    """
    Create a relion star file header.

    names - Header names as list

    Returns:
    header string
    """
    header = []
    header.append('')
    header.append('data_')
    header.append('')
    header.append('loop_')
    for index, name in enumerate(names):
        header.append('{0} #{1}'.format(name, index+1))
    return '\n'.join(header)


def create_export_data(export_data, lines):
    """
    Write export data to file.

    export_data - Data to export.
    file_name - Name of the file to write to.

    Returns:
    In place modificaion of lines
    """
    row_string = []
    for value in export_data:
        if isinstance(value, int):
            row_string.append('{0: 7d}'.format(value))
        elif isinstance(value, float):
            row_string.append('{0: 14f}'.format(value))
        else:
            row_string.append('{0:s}'.format(value))
    lines.append('{0}\n'.format(' '.join(row_string)))


@tu.rerun_function_in_case_of_error
def create_jpg_file(input_file, data, settings):
    file_name = tu.get_name(input_file)
    data = data[0]

    tu.mkdir_p(os.path.join(settings['motion_folder_feedback_0'], 'jpg'))
    tu.mkdir_p(os.path.join(settings['motion_folder_feedback_0'], 'jpg_2'))
    tu.mkdir_p(os.path.join(settings['motion_folder_feedback_0'], 'json'))

    jpg_file_1 = os.path.join(settings['motion_folder_feedback_0'], 'jpg', '{0}.jpg'.format(file_name))
    jpg_file_2 = os.path.join(settings['motion_folder_feedback_0'], 'jpg_2', '{0}.jpg'.format(file_name))
    json_file = os.path.join(settings['motion_folder_feedback_0'], 'json', '{0}.json'.format(file_name))

    json_dict = {
        'label_x': 'Drift x / px',
        'label_y': 'Drift y / px',
        'is_equal': True,
        'data': [
            {
                'values_x': data[0].tolist(),
                'values_y': data[1].tolist(),
                'is_high_res': True,
                'label_plot': '',
                'marker': '',
                'color': None,
                },
            {
                'values_x': [data[0][0]],
                'values_y': [data[1][0]],
                'is_high_res': False,
                'label_plot': 'First',
                'marker': 'p',
                'color': None,
                },
            {
                'values_x': [data[0][-1]],
                'values_y': [data[1][-1]],
                'is_high_res': False,
                'label_plot': 'Last',
                'marker': 'd',
                'color': None,
                },
            ]
        }
    with open(json_file, 'w') as write:
        json.dump(json_dict, write, indent=1)

    arr_1 = None
    arr_2 = None
    try:
        with mrc.open(input_file) as mrc_file:
            input_data = mrc_file.data
    except ValueError:
        if os.path.splitext(input_file)[1] in ('.mrc', '.mrcs'):
            with mrc.open(input_file, 'r+', permissive=True) as mrc_file:
                mrc_file.header.map = mrc.constants.MAP_ID
        with mrc.open(input_file) as mrc_file:
            input_data = mrc_file.data
    if len(input_data.shape) == 3:
        input_data = np.sum(input_data, axis=0) / input_data.shape[0]
    input_data = input_data - np.mean(input_data)
    input_data = tu.normalize_image(input_data)

    idx = 1
    while True:
        if np.max(input_data.shape) < idx * 4096:
            break
        idx += 1
    original_shape = idx * 4096
    bin_shape = 512
    ratio = original_shape / bin_shape
    assert ratio.is_integer()
    ratio = int(ratio)
    pad_x = original_shape - input_data.shape[0]
    pad_y = original_shape - input_data.shape[1]

    input_data = np.pad(input_data, ((0, pad_x), (0, pad_y)), mode='median')
    shape = (bin_shape, bin_shape)
    output_data = tu.rebin(input_data, shape)[:-int(1+pad_x//ratio), :-int(1+pad_y//ratio)]
    arr_1 = output_data

    tile_overlap = 512 / 2
    tile_shape = 512
    tile_ratio = original_shape / tile_overlap
    assert tile_ratio.is_integer()
    tile_ratio = int(tile_ratio)
    tile_images = []
    for x_val in range(tile_ratio):
        if (x_val+1)*tile_shape > original_shape:
            continue
        for y_val in range(tile_ratio):
            if (y_val+1)*tile_shape > original_shape:
                continue
            slices = (
                slice(x_val*tile_shape, (x_val+1)*tile_shape, 1),
                slice(y_val*tile_shape, (y_val+1)*tile_shape, 1),
                )
            pw = np.abs(np.fft.fftshift(np.fft.fft2(input_data[slices])))**2
            tile_images.append(pw)
    if tile_images:
        arr_2 = np.sum(np.array(tile_images) / len(tile_images), axis=0)
        arr_2 = tu.rebin(arr_2, shape)

        apix = None
        motion_name = settings['Copy']['Motion']
        if motion_name.startswith('MotionCor2'):
            if tu.is_higher_version(motion_name, '1.0.0'):
                apix = float(settings[motion_name]['-PixSize'])
        elif motion_name.startswith('Unblur'):
            if tu.is_higher_version(motion_name, '1.0.0'):
                apix = float(settings[motion_name]['Pixel size of image (A)'])
        if apix is None:
            raise Exception('Could not get pixel size!')

        arr_2 = tu.normalize_image(arr_2, apix=apix, real=False)

    if arr_1 is not None:
        mi.imsave(jpg_file_1, arr_1, cmap='gist_gray')
    if arr_2 is not None:
        mi.imsave(jpg_file_2, arr_2, cmap='gist_gray')



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import numpy as np


def get_mic_number(array, settings, as_int=True):
    """
    Identify the micrograph number out of the name string.

    Arguments:
    array - Array containing information
    settings - User provided settings

    Return:
    Array of micrograph numbers
    """
    error = False
    if as_int:
        out_type = int
    else:
        out_type = str
    if settings['Output']['Rename micrographs'] == 'True':
        number_list = []
        prefix = settings['Output']['Rename prefix']
        suffix = settings['Output']['Rename suffix']
        for entry in array:
            entry = os.path.basename(entry)
            if suffix == '':
                first_part_split = entry.rsplit('.', 1)
            else:
                first_part_split = entry.rsplit(suffix, 1)
            first_part = first_part_split[0]
            if prefix == '':
                number = first_part
            else:
                number = first_part.split(prefix)[-1]
            if as_int:
                try:
                    number = int(number)
                except ValueError:
                    error = True
                    break
            number_list.append(number)
    else:
        number_list = np.arange(len(array))

    if error:
        number_list = np.arange(len(array))
    else:
        pass

    return np.array(number_list).astype(out_type)


def dummy(data, settings, label):
    return '', '', ['', ''], ['', ''], ''


def update_ctf(data, settings, label):
    """
    Update the plot for CTFFIND v4.1.8.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values = get_mic_number(data['file_name'], settings)
    y_values = data[label]
    if label == 'defocus':
        y_values = y_values / 10000
        label = 'Defocus / mum'
        title = 'Mean defocus'

    elif label == 'defocus_diff':
        y_values = y_values / 10000
        label = 'Defocus diff / mum'
        title = 'Defocus diff'

    elif label == 'astigmatism':
        label = 'Angle / degree'
        title = 'Azimuth of astigmatism'

    elif label == 'phase_shift':
        label = 'Phase shift / degree'
        title = 'Additional phase shift'

    elif label == 'cross_corr':
        label = 'Cross correlation'
        title = 'Cross correlation'

    elif label == 'limit':
        y_values[y_values == np.inf] = 0
        label = 'Resolution limit / A'
        title = 'Resolution limit'

    else:
        raise Exception('Plotwidget: Do not know what to do :O {0}'.format(label))

    return x_values, y_values, ['Micrograph', 'Nr. of Micrographs'], [label, label], title


def update_motion(data, settings, label):
    """
    Update the plot for MotionCor2 v1.0.0.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values = get_mic_number(data['file_name'], settings)
    y_values = data[label]
    label_y = 'Drift / px'
    if label == 'overall drift':
        title = 'Overall drift'

    elif label == 'average drift per frame':
        title = 'Average drift per frame'

    elif label == 'first frame drift':
        title = 'First frame drift'

    elif label == 'average drift per frame without first':
        title = 'Average drift per frame without first'

    else:
        print('Plotwidget: Do not know what to do :O', label)
        raise Exception

    return x_values, y_values, ['Micrograph', 'Nr. of Micrographs'], [label_y, label_y], title


def update_cryolo_v1_0_4(data, settings, label):
    """
    Update the plot for crYOLO v1.0.4.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    y_values = data[label]
    x_values = get_mic_number(data['file_name'], settings)
    if label == 'particles':
        title = label
        label_x = ['Micrograph', 'Nr. of Micrographs']
        label_val = label
        label_hist = label
    else:
        title = label
        label_val = 'Mean {}'.format(label)
        label_hist = label
        y_values = np.array([np.mean(entry) for entry in y_values])
        label_x = ['Micrograph', 'Nr. of Particles']
    return x_values, y_values, label_x, [label_val, label_hist], title

def update_micrograph(data, settings, label):
    """
    Update the plot for crYOLO v1.0.4.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values = get_mic_number(data['file_name'], settings)
    y_values = data[label]
    title = label
    return x_values, y_values, ['Micrograph', 'Nr. of Micrographs'], [label, label], title

def update_batch(data, settings, label):
    """
    Update the plot for crYOLO v1.0.4.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values = get_mic_number(data['file_name'], settings)
    y_values = data[label]
    title = label
    return x_values, y_values, ['Batch', 'Nr. of Batches'], [label, label], title




"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import os
import glob
import shutil
import numpy as np
import matplotlib
matplotlib.use('qt5agg')
import matplotlib.pyplot as plt
import matplotlib.image as img

from . import transphire_utils as tu

def get_extract_command(file_sum, file_box, file_ctf, output_dir, settings, queue_com, name):
    """
    Create the ctf command based on the ctf software.

    file_input - Input name of the file for ctf estimation
    new_name - Output file
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of process

    Returns:
    CTF command
    File to check vor validation if the process was successful
    """
    extract_name = settings['Copy']['Extract']
    command = None
    block_gpu = None
    gpu_list = None
    shell = None
    if 'WINDOW' in extract_name:
        if tu.is_higher_version(extract_name, '1.2'):
            command = create_window_1_2_command(
                extract_name=extract_name,
                file_sum=file_sum,
                file_box=file_box,
                file_ctf=file_ctf,
                output_dir=output_dir,
                settings=settings
                )
            check_files = []
            block_gpu = False
            gpu_list = []
            shell = True
        else:
            message = '\n'.join([
                'Version for {0}: Not known!'.format(extract_name),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
            raise IOError(message)

    else:
        message = '\n'.join([
            '{0}: Not known!'.format(extract_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    assert command is not None, 'command not specified: {0}'.format(extract_name)
    assert block_gpu is not None, 'block_gpu not specified: {0}'.format(extract_name)
    assert gpu_list is not None, 'gpu_list not specified: {0}'.format(extract_name)
    assert shell is not None, 'shell not specified: {0}'.format(extract_name)

    return command, check_files, block_gpu, gpu_list, shell


def create_window_1_2_command(extract_name, file_sum, file_box, file_ctf, output_dir, settings):
    """Create the WINDOW v1.2 command"""

    try:
        shutil.rmtree(output_dir)
    except FileNotFoundError:
        pass

    command = []
    command.append('rm -rf {0}'.format(output_dir))
    command.append(';')
    # Add SPHIRE to the PATH
    command.append("PATH=$(dirname $(which {0})):${{PATH}}".format(settings['Path'][extract_name]))
    # Start the program
    mic_number = tu.get_name(file_sum)
    command.append(settings['Path'][extract_name])
    command.append("'{0}'".format(os.path.join(os.path.dirname(file_sum), os.path.basename(file_sum).replace(mic_number, '*'))))
    command.append("'{0}'".format(os.path.join(os.path.dirname(file_box), os.path.basename(file_box).replace(mic_number, '*'))))
    command.append(file_ctf)
    command.append(output_dir)
    command.append('--selection_list={0}'.format(file_sum))
    ignore_list = []
    ignore_list.append('--skip_invert')
    ignore_list.append('--limit_ctf')
    ignore_list.append('--check_consistency')
    for entry in ignore_list:
        if settings[extract_name][entry] == 'True':
            command.append(entry)

    for key in settings[extract_name]:
        if key in ignore_list:
            continue
        elif settings[extract_name][key]:
            command.append('{0}={1}'.format(key, settings[extract_name][key]))
        else:
            continue

    if settings[settings['Copy']['Picking']]['--filament'] == 'True':
        command.append('--coordinates_format=cryolo_helical_segmented')
    else:
        command.append('--coordinates_format=cryolo')

    command.append(';')
    command.append('mkdir -p {0}'.format(os.path.join(output_dir, 'png')))
    command.append(';')
    command.append("PATH=$(dirname $(which {0})):${{PATH}}".format(settings['Path']['e2proc2d.py']))
    command.append(settings['Path']['e2proc2d.py'])
    file_name = os.path.basename(output_dir)
    command.append(os.path.join(output_dir, '{0}_ptcls.mrcs'.format(file_name)))
    command.append(os.path.join(output_dir, 'png', '{0}_ptcls.mrcs.png').format(file_name))
    command.append('--meanshrink=4')
    command.append('--unstacking')
    command.append('--outmode=uint16')
    return ' '.join(command)


def find_logfiles(root_path, settings, queue_com, name):
    """
    Find logfiles related to the produced Extract files.

    root_path - Root path of the file
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of process

    Returns:
    list of log files
    """
    log_files = None
    copied_log_files = None
    extract_name = settings['Copy']['Extract']
    if 'WINDOW' in extract_name:
        if tu.is_higher_version(extract_name, '1.2'):
            copied_log_files = []
            recursive_file_search(directory=root_path, files=copied_log_files)
            log_files = copied_log_files
    else:
        message = '\n'.join([
            '{0}: Not known!'.format(settings['Copy']['Extract']),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    assert log_files is not None
    assert copied_log_files is not None
    return log_files, copied_log_files


def recursive_file_search(directory, files):
    """
    Recursive file search function.
    """
    file_names = glob.glob('{0}*'.format(directory))
    for name in file_names:
        if os.path.isdir(name):
            recursive_file_search('{0}/'.format(name), files)
        elif name.endswith('.png'):
            pass
        else:
            files.append(name)

@tu.rerun_function_in_case_of_error
def create_jpg_file(file_name, output_dir):
    files = sorted(glob.glob(os.path.join(output_dir, file_name, 'png', '*')))
    dpi = 300
    if len(files) == 0:
        columns = 1
        rows = 1
        width = 1
        height = 1
    else:
        columns = np.sqrt(2*len(files))
        columns = int(columns+bool(columns % 2))
        columns += bool(columns % 2)
        rows = int(len(files) / columns + 1)

        if columns * rows < len(files):
            rows += 1
        image = img.imread(files[0])
        width = image.shape[0] * columns / dpi
        height = image.shape[1] * rows / dpi

    fig, ax = plt.subplots(rows, columns, subplot_kw={'frameon':False, 'adjustable': 'box', 'aspect': 'equal'})
    ax = np.atleast_1d(ax).ravel()
    for idx, ax_instance in enumerate(ax):
        try:
            ax_instance.imshow(img.imread(files[idx]), cmap='Greys_r')
        except IndexError:
            pass
        ax_instance.axis('off')

    fig.set_size_inches(width, height)
    aspect = width / height
    plt.subplots_adjust(wspace=0.05, hspace=0.05/aspect, top=1, bottom=0, left=0, right=1)
    tu.mkdir_p(os.path.join(output_dir, 'jpg'))
    plt.savefig(os.path.join(output_dir, 'jpg', '{0}.jpg'.format(file_name)), dpi=dpi, transparent=True, edgecolor=None) 
    plt.close(fig)


def get_particle_number(log_file, settings, queue_com, name):
    extract_name = settings['Copy']['Extract']
    with open(log_file, 'r') as read:
        if 'WINDOW' in extract_name:
            if tu.is_higher_version(extract_name, '1.2'):
                n_particles = re.search('Global.*Processed\s*:\s*(\d+)', read.read(), re.S).group(1) #https://regex101.com/r/Jyo5hi/1

            else:
                message = '\n'.join([
                    'Version for {0}: Not known!'.format(extract_name),
                    'Please contact the TranSPHIRE authors!'
                    ])
                queue_com['error'].put(
                    message,
                    name
                    )
                raise IOError(message)
        else:
            message = '\n'.join([
                '{0}: Not known!'.format(extract_name),
                'Please contact the TranSPHIRE authors!'
                ])
            queue_com['error'].put(
                message,
                name
                )
            raise IOError(message)
    return n_particles



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import glob
import os
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QScrollArea, QLabel
from PyQt5.QtCore import pyqtSlot, pyqtSignal

from . import transphire_utils as tu

class SettingsContainer(QWidget):
    """
    Widget for setting widgets.

    Inherits:
    QWidget
    """
    sig_change_use_movie = pyqtSignal(int)
    sig_adjust_tab = pyqtSignal(object, str)

    def __init__(self, content, name, global_dict, settings_folder, mount_worker, parent=None, **kwargs):
        """
        Initialise layout of the widget.

        Arguments:
        content - Content for the widget
        max_widgets - Maximum widgets per column
        parent - Parent widget (default None)

        Returns:
        None
        """
        super(SettingsContainer, self).__init__(parent)

        self.parent = parent
        self.name = name
        self.global_dict = None
        self.input_file_names = []
        try:
            content_others = kwargs['content_others']
        except KeyError:
            content_others = None
        try:
            content_mount = kwargs['content_mount']
            self.content_mount = {}
            for entry in content_mount:
                for widget in entry:
                    for key in widget:
                        if key == 'Mount name':
                            current_name = widget[key][0]
                        if key == 'Typ':
                            if widget[key][0] == 'Import':
                                self.content_mount[current_name] = os.path.join(settings_folder, current_name).replace(' ', '_')
 
        except KeyError:
            self.content_mount = None

        # TabDocker widget for Main and Advanced
        my_tab_docker = TabDocker(self)
        my_tab_docker.setTabPosition('South')
        my_tab_docker.tab_widget.setObjectName('bot')
        layout_main = QVBoxLayout(self)
        layout_main.setContentsMargins(0, 0, 0, 0)

        self.layout_dict = {
            'Main_count': 0,
            'Advanced_count': 0,
            'Rare_count': 0,
            }
        important_note = ''
        for entry in content:
            for widget in entry:
                for key in widget:
                    if key == 'WIDGETS MAIN':
                        self.layout_dict['Main_max'] = int(widget[key][0])
                    elif key == 'WIDGETS ADVANCED':
                        self.layout_dict['Advanced_max'] = int(widget[key][0])
                    elif key == 'WIDGETS RARE':
                        self.layout_dict['Rare_max'] = int(widget[key][0])
                    elif key == 'IMPORTANT':
                        important_note = widget[key][0]
                    else:
                        self.layout_dict['{}_count'.format(widget[key][1]['widget_2'])] += 1
                        continue

        if important_note:
            important_widget = QLabel(important_note)
            important_widget.setObjectName('important')
            layout_main.addWidget(important_widget)
        layout_main.addWidget(my_tab_docker)

        for dict_name in ['Main', 'Advanced', 'Rare']:
            widget = QWidget(self)
            widget.setObjectName('settings')

            self.layout_dict[dict_name] = QHBoxLayout(widget)
            self.layout_dict[dict_name].setContentsMargins(3, 3, 3, 3)
            self.layout_dict[dict_name].addStretch(1)

            self.layout_dict['{0}_v'.format(dict_name)] = None
            self.layout_dict['{0}_idx'.format(dict_name)] = 0

            scroll_area = QScrollArea(my_tab_docker)
            scroll_area.setWidgetResizable(True)
            scroll_area.setWidget(widget)
            my_tab_docker.add_tab(scroll_area, '{}: {}'.format(dict_name, self.layout_dict['{}_count'.format(dict_name)]))

        # Global content
        self.content = {}
        self.group = {}

        # Add to layout
        for entry in content:
            for widget in entry:
                for key in widget:
                    if key == 'WIDGETS MAIN' or key == 'WIDGETS ADVANCED' or key == 'WIDGETS RARE' or key == 'IMPORTANT':
                        continue
                    layout_name = widget[key][1]['widget_2']
                    widget_name = widget[key][1]['name']
                    group = widget[key][1]['group']

                    if self.layout_dict['{0}_idx'.format(layout_name)] % self.layout_dict['{0}_max'.format(layout_name)] == 0:
                        if self.layout_dict['{0}_v'.format(layout_name)] is not None:
                            self.layout_dict[layout_name].addWidget(Separator(typ='vertical', color='lightgrey'))
                            self.layout_dict['{0}_v'.format(layout_name)].addStretch(1)
                        else:
                            pass
                        self.layout_dict['{0}_v'.format(layout_name)] = QVBoxLayout()
                        self.layout_dict['{0}_v'.format(layout_name)].setContentsMargins(0, 0, 0, 0)
                        self.layout_dict[layout_name].addLayout(self.layout_dict['{0}_v'.format(layout_name)])

                    settings_widget = SettingsWidget(
                        content=widget[key],
                        name=name,
                        content_others=content_others,
                        mount_directory=mount_worker.mount_directory,
                        global_dict=global_dict,
                        input_file_names=self.input_file_names,
                        parent=self
                        )

                    if self.name == 'Copy':
                        try:
                            settings_widget.edit.currentTextChanged.connect(self.prepare_send_adjust)
                        except AttributeError:
                            pass

                    if widget[key][1]['name'] == 'Use movies':
                        try:
                            settings_widget.edit.currentIndexChanged.connect(self.sig_change_use_movie.emit)
                        except AttributeError:
                            pass

                    if group and name not in ('Pipeline'):
                        group, state = group.split(':')
                        self.group.setdefault(group, [])
                        self.group[group].append([settings_widget, state, widget_name])
                    self.content[widget_name] = settings_widget
                    self.layout_dict['{0}_v'.format(layout_name)].addWidget(settings_widget)
                    self.layout_dict['{0}_idx'.format(layout_name)] += 1

        for key in self.group:
            self.content[key].sig_index_changed.connect(self.change_state)
            self.change_state(name=key)

        for dict_name in ['Main', 'Advanced', 'Rare']:
            try:
                self.layout_dict['{0}_v'.format(dict_name)].addStretch(1)
                self.layout_dict[dict_name].addStretch(1)
            except AttributeError:
                pass

    def emit_signals(self):
        for key in self.content:
            try:
                self.content[key].edit.currentTextChanged.emit(
                    self.content[key].edit.currentText()
                    )
            except AttributeError:
                pass

    @pyqtSlot(str)
    def prepare_send_adjust(self, text):
        self.sig_adjust_tab.emit(self.sender(), text)

    @pyqtSlot(str)
    def change_state(self, name):
        """
        Change the state of widgets based on the choice of the corresponding combo box

        name - Name of the group to change status (Emitted by the combo box)

        Returns:
        None
        """
        try:
            for entry in self.group[name]:
                widget = entry[0]
                state = entry[1]
                sub_name = entry[2]
                if not self.content[name].isEnabled():
                    widget.setEnabled(False)
                elif self.content[name].edit.currentText() == state:
                    widget.setEnabled(True)
                else:
                    widget.setEnabled(False)
                self.change_state(name=sub_name)
        except KeyError:
            return None

    def get_input_names(self):
        return self.input_file_names

    def get_settings(self, quiet=False):
        """
        Get the settings as dict

        Arguments:
        quiet - If True, no prints are executed

        Returns:
        Settings as dictionary
        """
        settings = {}
        error = False
        for key in self.content:
            dictionary = self.content[key].get_settings(quiet=quiet)
            if dictionary is None:
                error = True
            else:
                settings.update(dictionary)
            if self.name == 'Copy':
                new_key = '{0}_entries'.format(key.replace(' ', '_'))
                try:
                    entries = self.content[key].get_combo_entries()
                except AttributeError:
                    pass
                else:
                    test_list = ['False', 'True']
                    if set(test_list) == set(entries):
                        continue
                    elif 'Symlink' in entries:
                        continue
                    settings[new_key] = self.content[key].get_combo_entries()
            else:
                pass
        if error:
            return None
        else:
            return [settings]

    def set_design(self, settings):
        """
        Set settings to the widgets

        Arguments:
        settings - Settings to set.

        Returns:
        None
        """

        # Clear old layout design
        for dict_name in ['Main', 'Advanced', 'Rare']:
            self.layout_dict['{0}_max'.format(dict_name)] = int(settings['WIDGETS {0}'.format(dict_name.upper())])
            self.layout_dict['{0}_v'.format(dict_name)] = None
            self.layout_dict['{0}_idx'.format(dict_name)] = 0
            for idx in reversed(range(self.layout_dict[dict_name].count())):
                item = self.layout_dict[dict_name].itemAt(idx)
                self.layout_dict[dict_name].removeItem(item)
                if isinstance(item, QVBoxLayout):
                    for idx in reversed(range(item.count())):
                        item2 = item.itemAt(idx)
                        item.removeItem(item2)
                    item.setParent(None)
                else:
                    try:
                        item.widget().setParent(None)
                    except AttributeError:
                        pass
            self.layout_dict[dict_name].addStretch(1)

        for key in settings:
            if key == 'WIDGETS MAIN' or key == 'WIDGETS ADVANCED' or key == 'WIDGETS RARE':
                continue
            layout_name = settings[key]
            try:
                widget = self.content[key]
            except KeyError:
                if self.name == 'Path':
                    continue
                elif key == 'IMPORTANT':
                    continue
                else:
                    raise

            if self.layout_dict['{0}_idx'.format(layout_name)] % self.layout_dict['{0}_max'.format(layout_name)] == 0:
                if self.layout_dict['{0}_v'.format(layout_name)] is not None:
                    self.layout_dict[layout_name].addWidget(Separator(typ='vertical', color='lightgrey'))
                    self.layout_dict['{0}_v'.format(layout_name)].addStretch(1)
                else:
                    pass
                self.layout_dict['{0}_v'.format(layout_name)] = QVBoxLayout()
                self.layout_dict['{0}_v'.format(layout_name)].setContentsMargins(0, 0, 0, 0)
                self.layout_dict[layout_name].addLayout(self.layout_dict['{0}_v'.format(layout_name)])

            self.layout_dict['{0}_v'.format(layout_name)].addWidget(widget)
            self.layout_dict['{0}_idx'.format(layout_name)] += 1

        for dict_name in ['Main', 'Advanced', 'Rare']:
            try:
                self.layout_dict['{0}_v'.format(dict_name)].addStretch(1)
                self.layout_dict[dict_name].addStretch(1)
            except AttributeError:
                pass

    def set_settings(self, settings):
        """
        Set settings to the widgets

        Arguments:
        settings - Settings to set.

        Returns:
        None
        """
        for key in settings:
            if key.endswith('_global'):
                continue
            try:
                content = self.content[key]
            except KeyError:
                try: # This block has been introduced for backwards compatibility changes.
                    content = self.content[key.replace(' v', ' >=v')]
                except KeyError:
                    if self.name == 'Copy' and key.endswith('_entries'):
                        continue
                    else:
                        #print('Content for {0} is disabled.'.format(key))
                        continue

            try:
                key_global = '{0}_global'.format(key)
                global_settings = settings[key_global] if key_global in settings else None
                content.set_settings(settings[key], global_settings)
            except KeyError:
                if self.name == 'Copy' and key.endswith('_entries'):
                    continue
                else:
                    print(
                        'Setting changed: {0}!'.format(key),
                        'Please do provide new values'
                        )

    def enable(self, var, use_all):
        """
        Disable or enable all widgets

        Arguments:
        var - If True, enable the widgets, else disable them
        use_all - If True, all widgets are enabled/disabled

        Returns:
        None
        """
        for key in self.content:
            if use_all:
                self.content[key].setEnabled(var)
            else:
                self.content[key].setEnabled(var)

        if var:
            try:
                self.set_settings(self.get_settings(quiet=True)[0])
            except TypeError:
                pass
        else:
            pass

    def set_global(self, global_dict):
        self.global_dict = global_dict
        for key in self.content:
            try:
                self.content[key].edit.textChanged.connect(self.update_global)
            except AttributeError:
                self.content[key].edit.currentTextChanged.connect(self.update_global)
        self.emit_global()

    @pyqtSlot(int)
    def emit_global(self, _=None):
        for key in self.content:
            try:
                self.content[key].edit.textChanged.emit(self.content[key].edit.text())
            except AttributeError:
                self.content[key].edit.currentTextChanged.emit(self.content[key].edit.currentText())

    @pyqtSlot(str, str)
    def set_new_model(self, weights, threshold):
        self.content['--weights'].edit.setText(weights)
        self.content['--confidence_threshold'].edit.setText(threshold)

    @pyqtSlot(str)
    def update_global(self, text):
        if self.sender().parent().name not in self.global_dict:
            return

        for entry in self.global_dict[self.sender().parent().name]:

            is_movie_mode = None
            try:
                is_movie_mode = entry.parent.content['Use movies'].get_settings()['Use movies'] == 'True'
            except KeyError:
                pass

            if is_movie_mode is None:
                pass
            elif is_movie_mode and self.sender().parent().name == 'Pixel size bin':
                continue
            elif not is_movie_mode and self.sender().parent().name == 'Pixel size':
                continue

            entry.global_value = text
            try:
                if entry.widget_auto.isChecked():
                    try:
                        entry.edit.setText(text)
                    except AttributeError:
                        entry.edit.setCurrentText(text)

                    entry.widget_auto.toggled.emit(True)
            except AttributeError:
                pass


    def search_for_projects(self, project_dir):
        text = self.content['Software'].get_settings()['Software']
        project_dir = os.path.realpath(project_dir)

        if text.startswith('EPU'):
            if not [entry for entry in os.listdir(project_dir) if entry.startswith('Images-Disc')]:
                tu.message('Could not find "Images-Disc" folder in the specified directory! Please check if you specified the correct project directory created by EPU or your EPU version is correct.')
                return


            mount_path = None
            for key in self.content_mount:
                if os.stat(self.content_mount[key]).st_size == 0:
                    continue
                with open(self.content_mount[key], 'r') as read:
                    current_mount_path = os.path.realpath(read.readline().split('\t')[2])
                if project_dir.startswith(current_mount_path):
                    mount_path = current_mount_path
                    break

            if mount_path is None:
                tu.message('Could not identify related mounted mount point. Please provide the "Input project path for frames", "Input project path for jpg" and "Input frames extension" manually.')
                return

            matches = []
            self.recursive_search(mount_path, os.path.basename(project_dir), matches)

            if len(matches) in (1, 2):
                pass
            elif not matches:
                tu.message('Could not find project {0} in specified mount directory {1}. Please provide the "Input project path for frames", "Input project path for jpg" and "Input frames extension" manually.'.format(project_dir, mount_path))
                return
            else:
                tu.message('Found more than 2 folders with the name {0} in specified mount directory {1}. Please provide the "Input project path for frames", "Input project path for jpg" and "Input frames extension" manually.'.format(project_dir, mount_path))
                return

            is_meta = []
            is_frames = []
            extension = []
            possible_extensions = self.content['Input frames extension'].get_combo_entries()
            for entry in matches:
                files = glob.glob(os.path.join(entry, 'Images-Disc*', '*', 'Data', '*'))

                for file_name in files:
                    if file_name.endswith('.jpg'):
                        is_meta.append(entry)

                    if 'ractions.' in file_name:
                        is_frames.append(entry)
                        extension.extend(list(filter(
                            lambda x: file_name.endswith(x),
                            possible_extensions
                            )))

            if len(set(is_meta)) != 1 or len(set(is_frames)) != 1:
                tu.message('Could not identify meta and frames folder, yet. Did the data collection already start? If not, try again afterwards :)')
                return

            self.content['Input project path for frames'].set_settings(is_frames[0], '[None, None]')
            self.content['Input project path for jpg'].set_settings(is_meta[0], '[None, None]')

            if len(set(extension)) != 1:
                tu.message('Input project path for frames:\n{0}\n\nInput project path for jpg:\n{1}\n\Input frames extension:\nFOUND SEVERAL - Please provide manually!'.format(is_frames[0], is_meta[0]))
            else:
                self.content['Input frames extension'].set_settings(extension[0], '[None, None]')
                tu.message('Input project path for frames:\n{0}\n\nInput project path for jpg:\n{1}\n\nInput frames extension:\n{2}'.format(is_frames[0], is_meta[0], extension[0]))

        elif text.startswith('Just Stack'):
            self.content['Input project path for frames'].set_settings(project_dir, '[None, None]')
            self.content['Input project path for jpg'].set_settings(project_dir, '[None, None]')
            self.content['Number of frames'].set_settings('-1', '[None, None]')

            possible_extensions = self.content['Input frames extension'].get_combo_entries()
            extension = []
            for file_name in glob.iglob(os.path.join(project_dir, '*')):
                extension.extend(list(filter(
                    lambda x: file_name.endswith(x),
                    possible_extensions
                    )))

            if len(set(extension)) != 1:
                tu.message('Input project path for frames:\n{0}\n\nInput project path for jpg:\n{1}\n\nNumber of frames:\n-1\n\nInput frames extension:\nFOUND SEVERAL - Please provide manually!'.format(project_dir, project_dir))
            else:
                self.content['Input frames extension'].set_settings(extension[0], '[None, None]')
                tu.message('Input project path for frames:\n{0}\n\nInput project path for jpg:\n{1}\n\nNumber of frames:\n-1\n\nInput frames extension:\n{2}'.format(project_dir, project_dir, extension[0]))

        else:
            tu.message('Automatic folder detection is currently only supported for the EPU software. Please provide the "Input project path for frames", "Input project path for jpg" and "Input frames extension" manually.')

    @staticmethod
    def recursive_search(folder, match, matches):
        folders = sorted([entry for entry in glob.glob(os.path.join(folder, '*')) if os.path.isdir(entry)])
        for folder_name in folders:
            if os.path.basename(folder_name) == match:
                matches.append(folder_name)
                continue
            elif list(filter(
                    lambda x: os.path.basename(folder_name).startswith(x),
                    ('Images-Disc', 'Metadata', 'Sample', 'Atlas')
                    )):
                continue
            else:
                SettingsContainer.recursive_search(folder_name, match, matches)

from .settingswidget import SettingsWidget
from .separator import Separator
from .tabdocker import TabDocker



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import subprocess
import os
import glob
import re
import numpy as np

from . import transphire_utils as tu


def get_header(input_file):
    """
    Extract header information from a relion star file.

    Arguments:
    input_file - Input star file

    Return:
    Color string
    """
    with open(input_file, 'r') as read:
        lines = read.readlines()
    header = []
    idx = None
    for idx, line in enumerate(lines):
        line = line.lstrip().rstrip()
        if line.startswith('_rln'):
            name = line.split()[0]
            name = name.replace('_rlnFinalResolution', '_rlnCtfMaxResolution')
            if name == '_rlnMicrographName' or \
                    name == '_rlnCtfImage':
                header.append((name, '|U1200'))
            else:
                header.append((name, '<f8'))
        elif line == '' or line == 'data_' or line == 'loop_':
            continue
        else:
            break

    if idx is None:
        raise ValueError
    else:
        return header, idx


def get_dtype_dict():
    """
    Dtype of the data plot array.

    Arguments:
    None

    Return:
    Dtype dict
    """
    dtype = {}
    dtype['Motion'] = [
        ('overall drift', '<f8'),
        ('average drift per frame', '<f8'),
        ('first frame drift', '<f8'),
        ('average drift per frame without first', '<f8'),
        ('file_name', '|U1200'),
        ('image', '|U1200'),
        ]

    dtype['CTF'] = [
        ('mic_number', '<f8'),
        ('defocus', '<f8'),
        ('defocus_diff', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|U1200'),
        ('image', '|U1200'),
        ]

    dtype['Picking'] = [
        ('confidence', 'O'),
        ('box_x', 'O'),
        ('box_y', 'O'),
        ('particles', '<i8'),
        ('file_name', '|U1200'),
        ('image', '|U1200'),
        ]

    dtype['Extract'] = [
        ('accepted', '<i8'),
        ('rejected', '<i8'),
        ('file_name', '|U1200'),
        ('image', '|U1200'),
        ]

    dtype['Class2d'] = [
        ('classes', '<i8'),
        ('accepted', '<i8'),
        ('rejected', '<i8'),
        ('file_name', '|U1200'),
        ('image', '|U1200'),
        ]

    dtype['Train2d'] = [
        ('loss', '<f8'),
        ('file_name', '|U1200'),
        ]

    dtype['Auto3d'] = [
        ('resolution', '<f8'),
        ('file_name', '|U1200'),
        ('image', '|U1200'),
        ]

    dtype['Select2d'] = [
        ('accepted', '<i8'),
        ('accepted_percent', '<i8'),
        ('particles_accepted', '<i8'),
        ('particles_accepted_percent', '<i8'),
        ('rejected', '<i8'),
        ('rejected_percent', '<i8'),
        ('particles_rejected', '<i8'),
        ('particles_rejected_percent', '<i8'),
        ('file_name', '|U1200'),
        ('image', '|U1200'),
        ]

    dtype['Gctf >=v1.06'] = [
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|U1200')
        ]

    dtype['CTER >=v1.0'] = [
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|U1200')
        ]

    dtype['CTFFIND4 >=v4.1.8'] = [
        ('mic_number', '<f8'),
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|U1200')
        ]

    dtype['crYOLO >=v1.0.4'] = [
        ('coord_x', '<f8'),
        ('coord_y', '<f8'),
        ('box_x', '<f8'),
        ('box_y', '<f8'),
        ('file_name', '|U1200'),
        ]
    return dtype


def get_transphire_dict():
    """
    Translate transphire ctf dict into relion star file information.

    Arguments:
    None

    Return:
    Dtype dict
    """
    transphire_dict = {}
    transphire_dict = {
        'defocus_1': '_rlnDefocusU',
        'defocus_2': '_rlnDefocusV',
        'defocus': '_rlnDefocusU',
        'defocus_diff': '_rlnDefocusV',
        'astigmatism': '_rlnDefocusAngle',
        'phase_shift': '_rlnPhaseShift',
        'cross_corr': '_rlnCtfFigureOfMerit',
        'limit': '_rlnCtfMaxResolution',
        'file_name': '_rlnMicrographName',
        }
    return transphire_dict


def get_relion_dict():
    """
    Translate relion star file information to dtype dict.

    Arguments:
    None

    Return:
    Dtype dict
    """
    relion_dict = {}
    relion_dict = {
        '_rlnDefocusU': 'defocus_1',
        '_rlnDefocusV': 'defocus_2',
        '_rlnDefocusAngle': 'astigmatism',
        '_rlnPhaseShift': 'phase_shift',
        '_rlnCtfFigureOfMerit': 'cross_corr',
        '_rlnCtfMaxResolution': 'limit',
        '_rlnMicrographName': 'file_name',
        }
    return relion_dict


def get_dtype_import_dict():
    """
    Dtype of the file to import.

    Arguments:
    None

    Return:
    Dtype dict
    """
    dtype_import = {}
    dtype_import['CTER >=v1.0'] = [
        ('defocus', '<f8'),
        ('cs', '<f8'),
        ('volt', '<f8'),
        ('apix', '<f8'),
        ('bfac', '<f8'),
        ('amplitude_contrast', '<f8'),
        ('astigmatism_amplitude', '<f8'),
        ('astigmatism_angle', '<f8'),
        ('standard_deviation_defocus', '<f8'),
        ('standard_deviation_amplitude_contrast', '<f8'),
        ('standard_deviation_astigmatism_amplitude', '<f8'),
        ('standard_deviation_astigmatism_angle', '<f8'),
        ('coefficient_of_variation_of_defocus', '<f8'),
        ('coefficient_of_astigmatism_amplitude', '<f8'),
        ('limit_defocus', '<f8'),
        ('limit_defocus_and_astigmatism', '<f8'),
        ('limit_pixel_error', '<f8'),
        ('limit_maximum', '<f8'),
        ('reserved_spot', '<f8'),
        ('const_amplitude_contrast', '<f8'),
        ('phase_shift', '<f8'),
        ('file_name', '|U1200'),
        ]

    dtype_import['Gctf >=v1.06'] = [
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8')
        ]

    dtype_import['CTFFIND4 >=v4.1.8'] = [
        ('mic_number', '<f8'),
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8')
        ]

    dtype_import['MotionCor2 >=v1.0.0'] = [
        ('frame_number', '<f8'),
        ('shift_x', '<f8'),
        ('shift_y', '<f8')
        ]

    dtype_import['Unblur >=v1.0.0'] = [
        ('frame_number', '<i8'),
        ('shift_x', '<f8'),
        ('shift_y', '<f8')
        ]

    dtype_import['crYOLO >=v1.0.4'] = [
        ('coord_x', '<f8'),
        ('coord_y', '<f8'),
        ('box_x', '<f8'),
        ('box_y', '<f8'),
        ]
    return dtype_import


def dummy(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    if send_data is None:
        return None, None
    else:
        send_data.send((None, None))


def import_isac_v1_2(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    files = [
        entry for entry in glob.glob(
        '{0}/*/ISAC2'.format(directory_name)
        )
        ]
    useable_files = []
    for file_name in files:
        try:
            with open(os.path.join(file_name, 'processed_images.txt'), 'r') as read:
                accepted = len([entry for entry in read.readlines() if entry.strip()])
        except FileNotFoundError:
            accepted = 0
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        try:
            with open(os.path.join(file_name, 'not_processed_images.txt'), 'r') as read:
                rejected = len([entry for entry in read.readlines() if entry.strip()])
        except FileNotFoundError:
            rejected = 0
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        classes = len([entry for entry in glob.glob(
            '{0}/png/*'.format(os.path.dirname(file_name))
            )])
        useable_files.append([os.path.dirname(file_name), accepted, rejected, classes])

    useable_files_jpg = [
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ]

    useable_files = [
        [entry[0], entry[1], entry[2], entry[3]]
        for entry in sorted(useable_files)
        if tu.get_name(entry[0]) in useable_files_jpg
        ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['Class2d']
        )
    data = np.atleast_1d(data)
    data.fill(0)

    file_names_jpg = [tu.get_name(entry[0]) for entry in useable_files]
    jpgs = sorted([
        os.path.basename(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*'))
        ])
    jpg_names = [
        ';;;'.join([
            os.path.join(directory_name, jpg_dir_name, '{0}.jpg'.format(entry))
            for jpg_dir_name in jpgs
            ])
        for entry in file_names_jpg
        ]

    for idx, entry in enumerate(useable_files):
        data['file_name'][idx] = entry[0]
        data['accepted'][idx] = entry[1]
        data['rejected'][idx] = entry[2]
        data['classes'][idx] = entry[3]
    data['image'] = jpg_names

    data = np.sort(data, order='file_name')
    if send_data is None:
        return data, data
    else:
        send_data.send((data, data))


def import_cinderella_v0_3_1(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    files = [
        entry for entry in glob.glob(
        '{0}/{1}*_transphire.log'.format(directory_name, import_name)
        )
        ]
    useable_files = []
    for file_name in files:
        try:
            with open(file_name, 'r') as read:
                # Regex documentation can be found here: https://regex101.com/r/MxOgyg/3
                match = re.search(
                    '^\s*Good(?: classes|):\s*(\d+) .*$(?:\n|\r\n)(?:\n|\r\n)(?:\n|\r\n)^\s*Bad(?: classes|):\s*(\d+) .*$(?:\n|\r\n)(?:\n|\r\n)^Bad Particles(?:\n|\r\n)(\d+)(?:\n|\r\n)Good Particles(?:\n|\r\n)(\d+)$',
                    read.read(),
                    re.MULTILINE
                    )
        except FileNotFoundError:
            continue
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        if match is not None:
            useable_files.append([file_name, int(match.group(1)), int(match.group(2)), int(match.group(3)), int(match.group(4))])

    useable_files_jpg = [
        tu.get_name(entry).replace('_good', '').replace('_bad', '')
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ]

    useable_files = [
        [entry[0].replace('_transphire', ''), entry[1], entry[2], entry[3], entry[4]]
        for entry in sorted(useable_files)
        if tu.get_name(entry[0]).replace('_transphire', '') in useable_files_jpg
        ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['Select2d']
        )
    data = np.atleast_1d(data)
    data.fill(0)

    file_names_jpg = [tu.get_name(entry[0]) for entry in useable_files]
    jpgs = sorted([
        os.path.basename(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*'))
        ])
    jpg_names = [
        ';;;'.join([
            os.path.join(directory_name, jpg_dir_name, '{0}_bad.jpg'.format(entry))
            if idx == 0
            else
            os.path.join(directory_name, jpg_dir_name, '{0}_good.jpg'.format(entry))
            for idx, jpg_dir_name in enumerate(jpgs)
            ])
        for entry in file_names_jpg
        ]

    for idx, entry in enumerate(useable_files):
        data['file_name'][idx] = entry[0]
        data['accepted'][idx] = entry[1]
        data['rejected'][idx] = entry[2]
        data['accepted_percent'][idx] = 100 * entry[1] / (entry[1] + entry[2])
        data['rejected_percent'][idx] = 100 * entry[2] / (entry[1] + entry[2])
        data['particles_rejected'][idx] = entry[3]
        data['particles_accepted'][idx] = entry[4]
        data['particles_rejected_percent'][idx] = 100 * entry[3] / (entry[3] + entry[4])
        data['particles_accepted_percent'][idx] = 100 * entry[4] / (entry[3] + entry[4])
    data['image'] = jpg_names

    data = np.sort(data, order='file_name')
    if send_data is None:
        return data, data
    else:
        send_data.send((data, data))


def import_window_v1_2(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    files = [
        entry for entry in glob.glob(
        '{0}/{1}*_transphire.log'.format(directory_name, import_name)
        )
        ]
    useable_files = []
    for file_name in files:
        try:
            with open(file_name, 'r') as read:
                match = re.search(
                    '^.*Processed\s+:\s+(\d+).*$(?:\n|\r\n)^.*Rejected by out of boundary\s+:\s+(\d+).*$',
                    read.read(),
                    re.MULTILINE
                    )
        except FileNotFoundError:
            continue
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        if match is not None:
            useable_files.append([file_name, match.group(1), match.group(2)])

    useable_files_jpg = [
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ]

    useable_files = [
        [entry[0].replace('_transphire', ''), entry[1], entry[2]]
        for entry in sorted(useable_files)
        if tu.get_name(entry[0]).replace('_transphire', '') in useable_files_jpg
        ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['Extract']
        )
    data = np.atleast_1d(data)
    data.fill(0)

    file_names_jpg = [tu.get_name(entry[0]) for entry in useable_files]
    jpgs = sorted([
        os.path.basename(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*'))
        ])
    jpg_names = [
        ';;;'.join([
            os.path.join(directory_name, jpg_dir_name, '{0}.jpg'.format(entry))
            for jpg_dir_name in jpgs
            ])
        for entry in file_names_jpg
        ]

    for idx, entry in enumerate(useable_files):
        data['file_name'][idx] = entry[0]
        data['accepted'][idx] = entry[1]
        data['rejected'][idx] = entry[2]
    data['image'] = jpg_names

    data = np.sort(data, order='file_name')
    if send_data is None:
        return data, data
    else:
        send_data.send((data, data))


def import_ctffind_v4_1_8(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    """
    Import ctf information for CTFFIND v4.1.8.
    Defocus in angstrom, phase shift in degree.

    Arguments:
    name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    dtype_import_dict_name = tu.find_best_match(name_no_feedback, get_dtype_import_dict())
    dtype_dict_name = tu.find_best_match(name_no_feedback, get_dtype_dict())
    files = [
        entry for entry in glob.glob(
        '{0}/{1}*.txt'.format(directory_name, import_name)
        ) if not entry.endswith('_avrot.txt') and not '_transphire_' in entry
        ]

    useable_files = []
    for file_name in files:
        try:
            data_name = np.genfromtxt(
                file_name,
                dtype=get_dtype_import_dict()[dtype_import_dict_name],
                )
        except ValueError:
            continue
        except IOError:
            continue
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        else:
            if data_name.size > 0:
                useable_files.append([file_name, data_name])
            else:
                continue

    useable_files_jpg = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ])
    useable_files_json = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'json*', '*.json'))
        ])
    if not import_name: 
        useable_files = [
            entry
            for entry in sorted(useable_files)
            if tu.get_name(entry[0]) in useable_files_jpg and
            tu.get_name(entry[0]) in useable_files_json
            ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['CTF']
        )
    data_original = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()[dtype_dict_name]
        )
    data = np.atleast_1d(data)
    data_original = np.atleast_1d(data_original)
    data.fill(0)
    data_original.fill(0)

    match_re = re.compile('# Input file: (.*?)\s+; Number of micrographs: 1')

    file_names = []
    jpg_json_data = []
    jpg_dirs = glob.glob(os.path.join(directory_name, 'jpg*'))
    json_dirs = glob.glob(os.path.join(directory_name, 'json*'))
    for file_name, _ in useable_files:
        with open(file_name, 'r') as read:
            content = read.read()
        file_names.append(match_re.search(content, re.S).group(1))

        file_name_base = tu.get_name(file_name)
        jpgs = [os.path.join(jpg_name, '{}.jpg'.format(file_name_base)) for jpg_name in jpg_dirs]
        json = [os.path.join(json_name, '{}.json'.format(file_name_base)) for json_name in json_dirs]
        jpg_json_data.append(';;;'.join(jpgs + json))

    data_original['file_name'] = file_names
    data['file_name'] = file_names
    for dtype_name in data_original.dtype.names:
        if dtype_name == 'file_name':
            continue
        if dtype_name == 'phase_shift':
            data_original[dtype_name] = [np.degrees(entry[1][dtype_name]) for entry in useable_files]
        else:
            data_original[dtype_name] = [entry[1][dtype_name] for entry in useable_files]

        if dtype_name == 'defocus_1':
            data['defocus'] = [(entry[1]['defocus_2'] + entry[1]['defocus_1']) / 2 for entry in useable_files]
        elif dtype_name == 'defocus_2':
            data['defocus_diff'] = [entry[1]['defocus_2'] - entry[1]['defocus_1'] for entry in useable_files]
        elif dtype_name == 'phase_shift':
            data[dtype_name] = [np.degrees(entry[1][dtype_name]) for entry in useable_files]
        else:
            data[dtype_name] = [entry[1][dtype_name] for entry in useable_files]
    data['image'] = jpg_json_data

    data = np.sort(data, order='file_name')
    data_original = np.sort(data_original, order='file_name')
    if send_data is None:
        return data, data_original
    else:
        send_data.send((data, data_original))


def import_gctf_v1_06(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    """
    Import ctf information for Gctf v1.06.
    Defocus in angstrom, phase shift in degree.

    Arguments:
    name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    suffix = '_gctf'
    dtype_dict_name = tu.find_best_match(name_no_feedback, get_dtype_dict())

    useable_files = []
    for file_name in sorted(glob.glob('{0}/{1}*{2}.star'.format(directory_name, import_name, suffix))):
        try:
            dtype, max_header = get_header(input_file=file_name)
            data_name = np.genfromtxt(
                file_name,
                dtype=dtype,
                skip_header=max_header,
                )
        except ValueError:
            continue
        except IOError:
            continue
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        else:
            if data_name.size > 0:
                useable_files.append([file_name, data_name])
            else:
                continue

    useable_files_jpg = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ])
    useable_files_json = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'json*', '*.json'))
        ])
    if not import_name: 
        useable_files = [
            file_name
            for file_name in sorted(useable_files)
            if tu.get_name(file_name) in useable_files_jpg and
            tu.get_name(file_name) in useable_files_json
            ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['CTF']
        )
    data_original = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()[dtype_dict_name]
        )
    data = np.atleast_1d(data)
    data_original = np.atleast_1d(data_original)
    data.fill(0)
    data_original.fill(0)
    if not useable_files:
        if send_data is None:
            return None, None
        else:
            send_data.send((None, None))

    jpg_json_data = []
    for entry in sorted(useable_files):
        file_name_base = tu.get_name(file_name)
        jpgs = glob.glob(os.path.join(directory_name, 'jpg*', '{}.jpg'.format(file_name_base)))
        json = glob.glob(os.path.join(directory_name, 'json*', '{}.json'.format(file_name_base)))
        jpg_json_data.append(';;;'.join(jpgs + json))

    relion_dict = get_relion_dict()
    for dtype_name in useable_files[0][1].dtype.names:
        try:
            transphire_name = relion_dict[dtype_name]
        except KeyError:
            continue

        try:
            data_original[transphire_name] = np.nan_to_num([entry[1][dtype_name] for entry in useable_files])
        except ValueError:
            data_original[transphire_name] = 0

        if transphire_name == 'defocus_1':
            data['defocus'] = [(entry[1]['_rlnDefocusU']+entry[1]['_rlnDefocusV']) / 2 for entry in useable_files]
        elif transphire_name == 'defocus_2':
            data['defocus_diff'] = [entry[1]['_rlnDefocusV']-entry[1]['_rlnDefocusU'] for entry in useable_files]
        else:
            data[transphire_name] = [entry[1][dtype_name] for entry in useable_files]
            try:
                data[transphire_name][np.isinf(data[transphire_name])] = 0 # Set infinity to 0 to avoid histogram problems
            except TypeError:
                pass
            data[transphire_name] = np.nan_to_num(data[transphire_name], copy=False)

    data['image'] = jpg_json_data

    data = np.sort(data, order='file_name')
    data_original = np.sort(data_original, order='file_name')
    if send_data is None:
        return data, data_original
    else:
        send_data.send((data, data_original))


def import_cter_v1_0(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    """
    Import ctf information for CTER v1.0.
    Defocus in angstrom, phase shift in degree.

    Arguments:
    name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    dtype_import_dict_name = tu.find_best_match(name_no_feedback, get_dtype_import_dict())

    useable_files = []
    for file_name in sorted(glob.glob('{0}/{1}*/partres.txt'.format(directory_name, import_name))):
        try:
            data_name = np.genfromtxt(
                file_name,
                dtype=get_dtype_import_dict()[dtype_import_dict_name],
                )
        except ValueError:
            continue
        except IOError:
            continue
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        else:
            if data_name.size > 0:
                useable_files.append([os.path.dirname(file_name), data_name])
            else:
                continue

    useable_files_jpg = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ])
    useable_files_json = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'json*', '*.json'))
        ])
    if not import_name: 
        useable_files = [
            file_name
            for file_name in sorted(useable_files)
            if tu.get_name(file_name[0]) in useable_files_jpg and
            tu.get_name(file_name[0]) in useable_files_json
            ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['CTF']
        )
    data_original = np.zeros(
        len(useable_files),
        dtype=get_dtype_import_dict()[dtype_import_dict_name]
        )
    data = np.atleast_1d(data)
    data_original = np.atleast_1d(data_original)
    data.fill(0)
    data_original.fill(0)

    jpg_json_data = []
    for file_name in sorted(useable_files):
        file_name_base = tu.get_name(file_name[0])
        jpgs = glob.glob(os.path.join(directory_name, 'jpg*', '{}.jpg'.format(file_name_base)))
        json = glob.glob(os.path.join(directory_name, 'json*', '{}.json'.format(file_name_base)))
        jpg_json_data.append(';;;'.join(jpgs + json))


    for dtype_name in data_original.dtype.names:
        data_original[dtype_name] = [entry[1][dtype_name] for entry in useable_files]
        if dtype_name == 'defocus':
            data['defocus'] = [entry[1][dtype_name] * 10000 for entry in useable_files]
        elif dtype_name == 'astigmatism_amplitude':
            data['defocus_diff'] = [entry[1][dtype_name] * 10000 for entry in useable_files]
        elif dtype_name == 'astigmatism_angle':
            data['astigmatism'] = [45 - entry[1][dtype_name] for entry in useable_files]
        elif dtype_name == 'phase_shift':
            data['phase_shift'] = [entry[1][dtype_name] for entry in useable_files]
        elif dtype_name == 'file_name':
            data['file_name'] = [entry[1][dtype_name] for entry in useable_files]
        elif dtype_name == 'standard_deviation_defocus':
            data['cross_corr'] = [entry[1][dtype_name] for entry in useable_files]
        elif dtype_name == 'limit_defocus_and_astigmatism':
            data['limit'] = [1 / entry[1][dtype_name] if entry[1][dtype_name] != 0 else 1 / entry[1]['limit_pixel_error'] for entry in useable_files]
        else:
            continue
    data['image'] = jpg_json_data

    data = np.sort(data, order='file_name')
    data_original = np.sort(data_original, order='file_name')
    if send_data is None:
        return data, data_original
    else:
        send_data.send((data, data_original))


def import_motion_cor_2_v1_0_0(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    """
    Import motion information for MotionCor2 v1.0.0.

    Arguments:
    name - Name of motion program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    dtype_import_dict_name = tu.find_best_match(name_no_feedback, get_dtype_import_dict())

    directory_names = glob.glob('{0}/*_with*_DW_log'.format(directory_name))
    files = np.array(
        [
            entry
            for directory_name in directory_names
            for entry in glob.glob('{0}/{1}*-Full.log'.format(directory_name, import_name))
            ],
        dtype=str
        )

    useable_files = []
    for file_name in files:
        try:
            array = np.genfromtxt(
                file_name,
                dtype=get_dtype_import_dict()[dtype_import_dict_name]
                )
        except ValueError:
            continue
        except IOError:
            continue
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        else:
            if array.size > 0:
                useable_files.append(file_name)
            else:
                continue

    useable_files_jpg = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ])
    useable_files_json = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'json*', '*.json'))
        ])
    if not import_name: 
        useable_files = [
            file_name
            for file_name in sorted(useable_files)
            if tu.get_name(tu.get_name(file_name)) in useable_files_jpg and
            tu.get_name(tu.get_name(file_name)) in useable_files_json
            ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['Motion']
        )
    data = np.atleast_1d(data)
    data_original = []
    for idx, file_name in enumerate(useable_files):
        try:
            data_name = np.genfromtxt(
                file_name,
                dtype=get_dtype_import_dict()[dtype_import_dict_name]
                )
        except IOError:
            continue
        else:
            if data_name.size == 0:
                continue
            else:
                pass

        data_original.append([data_name['shift_x'], data_name['shift_y']])
        data[idx]['file_name'] = file_name.rsplit('0-', 1)[0]
        shift_x = np.array([
            data_name['shift_x'][i+1] - data_name['shift_x'][i] \
            for i in range(0, int(data_name['frame_number'][-1]-1))
            ])
        shift_y = np.array([
            data_name['shift_y'][i+1] - data_name['shift_y'][i] \
            for i in range(0, int(data_name['frame_number'][-1]-1))
            ])
        for entry in data.dtype.names:
            if entry == 'overall drift':
                data[idx][entry] = np.sum(np.sqrt(shift_x**2 + shift_y**2))
            elif entry == 'average drift per frame':
                data[idx][entry] = np.sum(np.sqrt(shift_x**2 + shift_y**2))/len(shift_x)
            elif entry == 'first frame drift':
                data[idx][entry] = np.sqrt(shift_x[0]**2 + shift_y[0]**2)
            elif entry == 'average drift per frame without first':
                data[idx][entry] = np.sum(np.sqrt(shift_x[1:]**2 + shift_y[1:]**2))/len(shift_x)
            else:
                pass

        jpg_name = os.path.join(
            directory_name,
            'jpg*',
            '{0}.jpg'.format(tu.get_name(tu.get_name(file_name)))
            )
        json_name = os.path.join(
            directory_name,
            'json*',
            '{0}.json'.format(tu.get_name(tu.get_name(file_name)))
            )
        data[idx]['image'] = ';;;'.join(glob.glob(jpg_name) + glob.glob(json_name))

    sort_idx = np.argsort(data, order='file_name')
    data = data[sort_idx]
    data_original = np.array(data_original)[sort_idx]
    if send_data is None:
        return data, data_original
    else:
        send_data.send((data, data_original))


def import_cryolo_v1_2_2(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    """
    Import picking information for crYOLO v1.2.2.

    Arguments:
    name - Name of picking program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    return import_cryolo_v1_0_4(
        name,
        name_no_feedback,
        settings,
        directory_name,
        sub_directory=['CBOX', 'EMAN', 'EMAN_HELIX_SEGMENTED'],
        import_name=import_name,
        send_data=send_data
        )


def import_cryolo_v1_0_4(name, name_no_feedback, settings, directory_name, import_name='', send_data=None, sub_directory=None, ):
    """
    Import picking information for crYOLO v1.0.4.

    Arguments:
    name - Name of picking program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """

    if sub_directory is None:
        sub_directory=['']
    box_files = []
    for dir_name in sub_directory:
        is_break = False
        for ext_name in ('cbox', 'box', 'txt'):
            box_files = glob.glob(os.path.join(
                directory_name,
                dir_name,
                '{0}*.{1}'.format(import_name, ext_name)
                ))
            if box_files:
                is_break = True
                break
        if is_break:
            break

    files_box = np.array(box_files)
    useable_files = []
    for file_name in files_box:
        try:
            data_imported = np.genfromtxt(file_name)
        except ValueError:
            useable_files.append([os.path.splitext(os.path.basename(file_name))[0], 0, np.array([0])])
        except IOError:
            continue
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        else:
            if file_name.endswith('.cbox') and data_imported.size != 0:
                data_cbox = np.atleast_2d(data_imported)[:, 4]
                data_box_x = np.atleast_2d(data_imported)[:, 5]
                data_box_y = np.atleast_2d(data_imported)[:, 6]
            else:
                data_cbox = np.array([0])
                data_box_x = np.array([0])
                data_box_y = np.array([0])
            useable_files.append([os.path.splitext(os.path.basename(file_name))[0], data_imported.shape[0], data_cbox, data_box_x, data_box_y])

    useable_files_jpg = [
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ]
    useable_files = [
        entry
        for entry in sorted(useable_files)
        if tu.get_name(entry[0]) in useable_files_jpg
        ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['Picking']
        )
    data = np.atleast_1d(data)
    file_names = [entry[0] for entry in useable_files]
    jpgs = sorted([os.path.basename(entry) for entry in glob.glob(os.path.join(directory_name, 'jpg*'))])
    jpg_names = [';;;'.join([os.path.join(directory_name, jpg_dir_name, '{0}.jpg'.format(entry)) for jpg_dir_name in jpgs]) for entry in file_names]
    data['file_name'] = file_names
    data['confidence'] = [entry[2] for entry in useable_files]
    data['box_x'] = [entry[3] for entry in useable_files]
    data['box_y'] = [entry[4] for entry in useable_files]
    data['particles'] = [entry[1] for entry in useable_files]
    data['image'] = jpg_names

    data_original = None

    data = np.sort(data, order='file_name')
    if send_data is None:
        return data, data_original
    else:
        send_data.send((data, data_original))


def import_unblur_v1_0_0(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    """
    Import motion information for cisTEM Unblur v1.0.0.

    Arguments:
    name - Name of motion program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    dtype_import_dict_name = tu.find_best_match(name_no_feedback, get_dtype_import_dict())

    directory_names = glob.glob('{0}/*_with*_DW_log'.format(directory_name))
    files = np.array(
        [
            entry
            for directory_name in directory_names
            for entry in glob.glob('{0}/{1}*_transphire.log'.format(directory_name, import_name))
            ],
        dtype=str
        )

    useable_files = []
    re_comp = re.compile('^image #(?P<frame>\d+) = (?P<xshift>[-\d.]+), (?P<yshift>[-\d.]+)$', re.M) # https://regex101.com/r/jmBPfH/1/ regex explanation
    for file_name in files:
        try:
            with open(file_name, 'r') as read:
                content = read.read()
        except ValueError:
            continue
        except IOError:
            continue
        except Exception as e:
            print('File corrupt: {} - {}'.format(file_name, str(e)))
        else:
            if len(re_comp.findall(content)) > 0:
                useable_files.append(file_name)
            else:
                continue

    useable_files_jpg = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'jpg*', '*.jpg'))
        ])
    useable_files_json = set([
        tu.get_name(entry)
        for entry in glob.glob(os.path.join(directory_name, 'json*', '*.json'))
        ])
    if not import_name: 
        useable_files = [
            file_name
            for file_name in sorted(useable_files)
            if tu.get_name(tu.get_name(file_name)) in useable_files_jpg and
            tu.get_name(tu.get_name(file_name)) in useable_files_json
            ]

    data = np.zeros(
        len(useable_files),
        dtype=get_dtype_dict()['Motion']
        )
    data = np.atleast_1d(data)
    data_original = []
    for idx, file_name in enumerate(useable_files):
        try:
            with open(file_name, 'r') as read:
                content = read.read()
        except ValueError:
            continue
        except IOError:
            continue
        else:
            matches = re_comp.findall(content)
            if len(matches) > 0:
                pass
            else:
                continue

        shift_x = []
        shift_y = []
        frame_list = []
        for match in matches:
            if int(match[0])+1 in frame_list:
                break
            shift_x.append(float(match[1]))
            shift_y.append(float(match[2]))
            frame_list.append(int(match[0])+1)

        data_name = np.empty(
            len(shift_x),
            dtype=get_dtype_import_dict()[dtype_import_dict_name]
            )
        data_name['shift_x'] = shift_x
        data_name['shift_y'] = shift_y
        data_name['frame_number'] = frame_list

        data_original.append([data_name['shift_x'], data_name['shift_y']])
        data[idx]['file_name'] = file_name
        shift_x = np.array([
            data_name['shift_x'][i+1] - data_name['shift_x'][i] \
            for i in range(0, int(data_name['frame_number'][-1]-1))
            ])
        shift_y = np.array([
            data_name['shift_y'][i+1] - data_name['shift_y'][i] \
            for i in range(0, int(data_name['frame_number'][-1]-1))
            ])
        for entry in data.dtype.names:
            if entry == 'overall drift':
                data[idx][entry] = np.sum(np.sqrt(shift_x**2 + shift_y**2))
            elif entry == 'average drift per frame':
                data[idx][entry] = np.sum(np.sqrt(shift_x**2 + shift_y**2))/len(shift_x)
            elif entry == 'first frame drift':
                data[idx][entry] = np.sqrt(shift_x[0]**2 + shift_y[0]**2)
            elif entry == 'average drift per frame without first':
                data[idx][entry] = np.sum(np.sqrt(shift_x[1:]**2 + shift_y[1:]**2))/len(shift_x)
            else:
                pass

        jpg_name = os.path.join(
            directory_name,
            'jpg*',
            '{0}.jpg'.format(tu.get_name(tu.get_name(file_name)))
            )
        json_name = os.path.join(
            directory_name,
            'json*',
            '{0}.json'.format(tu.get_name(tu.get_name(file_name)))
            )
        data[idx]['image'] = ';;;'.join(glob.glob(jpg_name) + glob.glob(json_name))

    sort_idx = np.argsort(data, order='file_name')
    data = data[sort_idx]
    data_original = np.array(data_original)[sort_idx]
    if send_data is None:
        return data, data_original
    else:
        send_data.send((data, data_original))


def import_auto_sphire_v1_3(name, name_no_feedback, settings, directory_name, import_name='', send_data=None):
    """
    Import motion information for auto_sphire.py version 1.3

    Arguments:
    name - Name of motion program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """

    mount_work = directory_name.replace(
        settings['Output']['Project directory'],
        settings['copy_to_work_folder_feedback_0'],
        )
    directory_names = glob.glob(os.path.join(mount_work, '{0}*_FILES'.format(import_name)))

    useable_files = []
    for entry in directory_names:
        if glob.glob(os.path.join(entry[:-6], '*_SHARPENING/vol_combined.hdf')):
            useable_files.append(entry[:-6])

    final_resolution = []
    final_jpg = []
    final_file_name = []
    for entry in useable_files:
        jpg_file = os.path.join(
            directory_name,
            'jpg',
            '{}.jpg'.format(os.path.basename(entry))
            )
        jpg_file_log = os.path.join(
            directory_name,
            'jpg',
            '{}.log'.format(os.path.basename(entry))
            )
        if not os.path.isfile(jpg_file):
            tu.mkdir_p(os.path.join(directory_name, 'jpg'))
            with open(jpg_file_log, 'w') as write:
                subprocess.call(
                    '{0} --script "{1}/support_scripts/chimerax.py {2} {3}" --nogui --offscreen'.format(
                        settings['Path']['chimerax'],
                        os.path.dirname(__file__),
                        glob.glob(os.path.join(entry, '*_SHARPENING/vol_combined.hdf'))[0],
                        jpg_file,
                        ),
                    shell=True,
                    stdout=write,
                    stderr=write,
                    )

        log_file = glob.glob(os.path.join(entry, '*_SHARPENING/log.txt'))[0]
        with open(log_file, 'r') as read:
            content = read.read()
        resolution = re.search('^.*FSC masked halves :.* 0\.143:\s*([.\d]*)A$', content, re.M).group(1) # https://regex101.com/r/6xdngz/1/

        final_resolution.append(float(resolution))
        final_jpg.append(';;;'.join([jpg_file]))
        final_file_name.append(tu.get_name(entry).replace('AUTOSPHIRE_', ''))

    data = np.zeros(
        len(final_jpg),
        dtype=get_dtype_dict()['Auto3d']
        )
    data = np.atleast_1d(data)
    data['resolution'] = final_resolution
    data['image'] = final_jpg
    data['file_name'] = final_file_name

    sort_idx = np.argsort(data, order='file_name')
    data = data[sort_idx]
    data_original = data
    if send_data is None:
        return data, data_original
    else:
        send_data.send((data, data_original))



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import os
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QLabel, QPushButton, QDialogButtonBox, QWidget, QLineEdit, QFileDialog, QCheckBox
from PyQt5.QtCore import pyqtSlot


class InputBox(QDialog):
    """
    Show a message box with an input field.

    Inherits:
    QDialog

    Signals:
    None
    """

    def __init__(self, is_password, restart_names=None, is_stop=False, settings=None, parent=None):
        """
        Initialise layout of the widget.

        Arguments:
        is_password - Input is a password and needs to be hidden.
        parent - Parent widget

        Returns:
        None
        """
        super(InputBox, self).__init__(parent)
        self.settings = settings

        if restart_names is None:
            restart_names = np.array([])
        central_raw_layout = QVBoxLayout(self)
        central_raw_layout.setContentsMargins(0, 0, 0, 0)
        central_widget_raw = QWidget(self)
        central_widget_raw.setObjectName('central_raw')
        central_raw_layout.addWidget(central_widget_raw)

        central_layout = QVBoxLayout(central_widget_raw)
        central_widget = QWidget(self)
        central_widget.setObjectName('central')
        central_layout.addWidget(central_widget)

        layout = QVBoxLayout(central_widget)
        self.label = QLabel(self)
        self.edit = QLineEdit('', self)
        self.edit.setObjectName('setting_xlarge')

        if is_password:
            self.edit.setEchoMode(QLineEdit.Password)
        else:
            pass

        button_box = QDialogButtonBox(self)
        accept_button = QPushButton('Ok')
        accept_button.setObjectName('start')
        reject_button = QPushButton('Cancel')
        reject_button.setObjectName('stop')

        button_box.addButton(accept_button, QDialogButtonBox.AcceptRole)
        button_box.addButton(reject_button, QDialogButtonBox.RejectRole)

        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        layout.addWidget(self.label)
        layout.addWidget(self.edit)
        layout.addStretch(1)

        if is_stop:
            self.abort = QCheckBox('Abort running processes?', self)
            self.abort.setChecked(True)
            layout.addWidget(self.abort)

        self.restart_content = {}
        if restart_names.size:
            layout.addWidget(QLabel('-----', self))
            self.restart_content['feedback'] = QCheckBox('Restart feedback - Check this if you modified "Number of feedbacks"', self)
            self.restart_content['feedback'].stateChanged.connect(self.handle_check)
            layout.addWidget(self.restart_content['feedback'])
            for entry in restart_names:
                if entry == 'Compress':
                    name = 'Restart {}'.format(entry)
                else:
                    name = 'Restart {} - Restarts subsequent jobs.'.format(entry)
                self.restart_content[entry] = QCheckBox(name, self)
                self.restart_content[entry].stateChanged.connect(self.handle_check)
                layout.addWidget(self.restart_content[entry])

        layout.addWidget(button_box)

    @pyqtSlot(int)
    def handle_check(self, state):
        ctf_name = self.settings['Copy']['CTF']
        is_movie = False
        try:
            is_movie = self.settings[ctf_name]['Use movies'] == 'True'
        except KeyError:
            pass
        sender = self.sender()
        key = sender.text().split()[1]
        if key not in ('Compress', 'feedback'):
            is_checked = None
            for input_key in self.restart_content:
                if key == input_key and is_checked is None:
                    cur_state = self.restart_content[input_key].blockSignals(True)
                    self.restart_content[input_key].setChecked(state)
                    self.restart_content[input_key].blockSignals(cur_state)
                    is_checked = 1 if state else 0
                elif key in ('Motion') and input_key in ('CTF') and is_movie:
                    continue
                elif key in ('CTF', 'Picking') and input_key in ('CTF', 'Picking'):
                    if not state and not is_checked:
                        is_checked = 1 if self.restart_content[input_key].isChecked() else 0
                elif key in ('Train2d', 'Auto3d') and input_key in ('Train2d', 'Auto3d'):
                    pass
                elif is_checked is not None:
                    cur_state = self.restart_content[input_key].blockSignals(True)
                    self.restart_content[input_key].setCheckState(is_checked)
                    self.restart_content[input_key].setEnabled(not is_checked)
                    self.restart_content[input_key].blockSignals(cur_state)
            if key == 'Motion' and not state:
                check_states = ['feedback']
                if is_movie:
                    check_states.append('CTF')
                for name in check_states:
                    if self.restart_content[name].isChecked():
                        self.restart_content[name].setCheckState(0)
                        self.restart_content[name].setCheckState(2)

        elif key == 'feedback':
            if self.restart_content['Picking'].isEnabled() or self.restart_content['Picking'].checkState() == 2:
                self.restart_content['Picking'].setCheckState(state)
                self.restart_content['Picking'].setEnabled(not state)

    def get_restart_dict(self):
        for key in self.restart_content:
            self.restart_content[key] = self.restart_content[key].checkState()
        return self.restart_content

    def setText(self, heading, text):
        """
        Set the text to the label.

        Arguments:
        heading - Heading of the window
        text - Text of the label

        Returns:
        None
        """
        if heading is not None:
            self.setWindowTitle(heading)
        self.label.setText(text)

    def getText(self):
        """
        Get the text from the label.

        Arguments:
        None

        Returns:
        Text content
        """
        return self.edit.text()

    def setDefault(self, text):
        """
        Set the default values for the QLineEdit.

        text - Text to put

        Returns:
        None
        """
        self.edit.setText(text)

    def set_type(self, this_type):
        if this_type in ('FILE', 'FILE/CHOICE'):
            self.edit.returnPressed.connect(self._find_file)
            self.edit.setToolTip(self.edit.toolTip() + '\n Shortcut: Shift + Return -> Open file dialog')
        elif this_type in ('DIR'):
            self.edit.returnPressed.connect(self._find_dir)
            self.edit.setToolTip(self.edit.toolTip() + '\n Shortcut: Shift + Return -> Open directory dialog')
        else:
            pass

    @pyqtSlot()
    def _find_file(self):
        """
        Find file with an open file dialog.

        Arguments:
        None

        Returns:
        None
        """
        in_file = QFileDialog.getOpenFileName(
            caption='Find file: {0}'.format(self.label.text()),
            directory=os.getcwd(),
            options=QFileDialog.DontUseNativeDialog
            )

        in_file = in_file[0]

        if in_file != '':
            self.sender().setText(in_file)


    @pyqtSlot()
    def _find_dir(self):
        """
        Find directory with an open directory dialog

        Arguments:
        None

        Returns:
        None
        """
        in_dir = QFileDialog.getExistingDirectory(
            caption='Find directory: {0}'.format(self.label.text()),
            directory=os.getcwd(),
            options=QFileDialog.DontUseNativeDialog
            )
        if in_dir != '':
            self.sender().setText(in_dir)

